#-------------------------------------------------------------------------------
#
# Copyright 2020 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
try:
    import zmq
    import time
    import threading
    import struct
    import clag.util as util
    import nlmanager.nlmanager
except ImportError as e:
    raise ImportError (str(e) + "- required module not found")

'''
Handles all interactions with CSU Manager
'''
class CSUClient:
    '''
    Manages Clagd communication with CS Manager
    '''
    ALL_MODID = 1
    CLAG_MODID = 6
    CSMGR_REQ = "tcp://127.0.0.1:49002"
    CSMGR_SUB = "tcp://127.0.0.1:49012"
    comm_socket = None
    REBOOT_COLD = 1
    MAINTENANCE_MODE = 0x100
    STATE_UP = 103
    STATE_DOWN = 101

    protoDownState = True
    pubSubThread = None
    monitorOperStatusThread = None
    mode = REBOOT_COLD
    state = STATE_UP

    keep_going = True

    def __init__(self, Log, Intf, Parser):
        self.Log = Log
        self.Intf = Intf
        self.Parser = Parser
        
        self.csuLock = threading.Lock()

        # Start a thread to handle publish/subscribe messages
        self.startPubSubThread()

        # setup communication socket with ZMQ
        self.initZmqComm()

        # Send load complete message
        self.sendLoadCompleteMsg()

        # Send Init complete message
        self.sendInitCompleteMsg()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_CSU, logStr)

    def log_verbose(self, logStr):
        self.Log.log_verbose(self.Log._LOG_DEBUG_CSU, logStr)

    def startPubSubThread(self):
        context = zmq.Context()
        socket = context.socket(zmq.SUB)
        socket.connect(self.CSMGR_SUB)
        socket.subscribe("")  # Subscribe to all topics
        time.sleep(1)        

        self.pubSubThread = threading.Thread(None, self.handlePubSubMesgs,
                                             "CS Manager PUB/SUB",
                                             args=(context, socket))
        self.pubSubThread.daemon = True
        self.pubSubThread.start()

    def initZmqComm(self):
        context = zmq.Context()
        self.comm_socket = context.socket(zmq.REQ)
        self.comm_socket.connect(self.CSMGR_REQ)

    def createLoadCompleteMsg(self):
        LOAD_COMPL_MSG = b"\x1c\x00\x00\x00\x03\x00\x00\x00\x18\x00\x00\x00\x06\x00\x00\x00\xc0\x78\xbd\x2a\x03\x00\x00\x00\x00\x00\x00\x00"
        return LOAD_COMPL_MSG

    def sendLoadCompleteMsg(self):
        TYPE_MODE_INFO = 104
        MODE_REBOOT_COLD = 1

        loadComplMsg = self.createLoadCompleteMsg()
        self.csuLock.acquire()
        self.comm_socket.send(loadComplMsg)

        resp = self.comm_socket.recv()

        # Parse the reply to get our mode and state
        msgLen = len(resp)
        index = 4
        if (msgLen <= 4):
            self.Log.log_error( "The msg received from CSU Manager has invalid mesg length")
            return

        lenInPkt = struct.unpack("i", resp[0:4])[0]
        if msgLen != lenInPkt:
            self.Log.log("received mesg length %d, doesn't match length in pkt %d" % ( msgLen, lenInPkt))
            return

        while (index < msgLen):
            bufferType = struct.unpack("i", resp[index:index+4])[0]
            bufferLen = struct.unpack("i", resp[index+4:index+8])[0]
            # buffer = msg[index]
            buffer = self.extractMessage(resp, index, index+bufferLen, msgLen)
            if (bufferType == TYPE_MODE_INFO):
                mod, mode, state = self.extractModuleStatus(buffer)
                self.setModeAndState(mode, state)
                if mode == MODE_REBOOT_COLD and self.isSlaveProtoDown() == True:
                    self.Log.log("CSU Cold Boot")
                    self.setSlavesProtoDown(False)

            index += bufferLen

        self.csuLock.release()
        # resp_hex = ' '.join(x.encode('hex') for x in resp)
        # print(resp_hex)

    def createInitCompleteMsg(self, mode):
        INIT_COMPL_MSG = b"\x1c\x00\x00\x00\x04\x00\x00\x00\x18\x00\x00\x00\x06\x00\x00\x00" + \
            mode.to_bytes(2, 'little') + b"\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00"
        return INIT_COMPL_MSG

    def sendInitCompleteMsg(self):
        initComplMsg = self.createInitCompleteMsg(self.mode)
        self.csuLock.acquire()
        self.comm_socket.send(initComplMsg)

        self.comm_socket.recv()
        self.csuLock.release()

    def createGoDownReplyMsg(self):
        DOWN_COMPL_MSG = b"\x20\x00\x00\x00\x64\x00\x00\x00\x1c\x00\x00\x00\x06\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        return DOWN_COMPL_MSG

    def sendGoDownReplyMsg(self):
        self.csuLock.acquire()
        goDownReplyMsg = self.createGoDownReplyMsg()
        self.comm_socket.send(goDownReplyMsg)
        self.log_debug("Sent GoDown reply")

        self.comm_socket.recv()
        self.csuLock.release()

    def createGoDownReplyOther(self, mod):
        DOWN_COMPL_OTHER_MSG = b"\x20\x00\x00\x00\x64\x00\x00\x00\x1c\x00\x00\x00" + \
            bytes([mod]) + b"\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        return DOWN_COMPL_OTHER_MSG

    def sendGoDownReplyOther(self, mod):
        self.csuLock.acquire()
        goDownReplyOtherMsg = self.createGoDownReplyOther(mod)
        self.comm_socket.send(goDownReplyOtherMsg)
        self.log_debug("Sending down for mod %d" % (mod))

        self.comm_socket.recv()
        self.csuLock.release()

    def createModuleStatusRespMsg(self, mod, mode, state):
        MODULE_STATUS_RESP_MSG = b"\x1c\x00\x00\x00\x06\x00\x00\x00\x18\x00\x00\x00" + \
            bytes([mod]) + b"\x00\x00\x00" + \
            mode.to_bytes(2, 'little') + b"\x00\x00" + \
            bytes([state]) + b"\x00\x00\x00" + b"\x00\x00\x00\x00"
        return MODULE_STATUS_RESP_MSG

    def sendModuleStatusRespMsg(self):
        self.csuLock.acquire()
        moduleStatusRespMsg = self.createModuleStatusRespMsg( self.CLAG_MODID, self.mode, self.state)
        self.comm_socket.send(moduleStatusRespMsg)
        self.log_debug("Sending ModuleStatusResp mode %d state %d" % (self.mode, self.state))

        self.comm_socket.recv()
        self.csuLock.release()

    def extractMessage(self, msg, start, end, totalLength):
        buffer = msg[start + 8: end] if (end < totalLength) else msg[start + 8:]
        return buffer

    def extractModMode(self, msg):
        mod = struct.unpack("i", msg[0:4])[0]
        mode = struct.unpack("i", msg[4:8])[0]
        state = struct.unpack("i", msg[8:])[0]

        return mod, mode, state

    def extractModuleStatus(self, msg):
        mod = struct.unpack("i", msg[0:4])[0]
        mode = struct.unpack("i", msg[4:8])[0]
        state = struct.unpack("i", msg[8:12])[0]

        return mod, mode, state

    # We don't do anything with KEEP_ALIVE_REQ for now. We could potentially
    # send back a response if that become mandatory later
    def handleKeepAliveReqMesg(self, msg):
        keepalive_seq = struct.unpack("i", msg)[0]
        self.log_verbose("keepalive seq =%d" % (keepalive_seq))

    def handleGoDownMesg(self, msg):
        mod, mode, state = self.extractModMode(msg)

        self.Log.log("GO_DOWN message received")
        self.log_debug("  mod=%d mode=%d state=%d" % (mod, mode, state))

        if mod == self.CLAG_MODID:
            self.setModeAndState(mode, state)
            self.setSlavesProtoDown(True)
        else:
            self.sendGoDownReplyOther(mod)

    def handleDownMesg(self, msg):
        mod, mode, state = self.extractModMode(msg)

        self.Log.log("DOWN message received")
        self.log_debug("  mod=%d mode=%d state=%d" % (mod, mode, state))

        self.setModeAndState(mode, state)

    def handleComeUpMesg(self, msg):
        mod, mode, state = self.extractModMode(msg)

        self.Log.log("COME_UP message received")
        self.log_debug("  mod=%d mode=%d state=%d" % (mod, mode, state))

        self.setModeAndState(mode, state)
        self.setSlavesProtoDown(False)
        self.sendLoadCompleteMsg()
        self.sendInitCompleteMsg()

    def handleUpMesg(self, msg):
        mod, mode, state = self.extractModuleStatus(msg)

        self.Log.log("UP message received")
        self.log_debug("  mod=%d mode=%d state=%d" % (mod, mode, state))

        self.setModeAndState(mode, state)

    def handleNetworkLayerInfoMesg(self, msg):
        rx_mod, rx_mode, rx_state = self.extractModuleStatus(msg)

        self.Log.log("Network layer info message received")
        self.log_debug("  mod=%d mode=%d state=%d" % (rx_mod, rx_mode, rx_state))

        # If we are in Maintenance mode OR If the state received in Network layer 
        # is down- ignore the message
        if self.mode & self.MAINTENANCE_MODE == self.MAINTENANCE_MODE or rx_state == self.STATE_DOWN:
            return

        if (self.isSlaveProtoDown() == True):
            self.setSlavesProtoDown(False)

    def handleModuleStatusReqMesg(self, msg):
        self.Log.log("Module Status Request message received")
        self.sendModuleStatusRespMsg()

    def handlePubSubMesgs(self, context, socket):
        self.Log.log("Thread to receive from CSU Manager -- Started")

        while self.keep_going:
            try:
                self.csuLock.acquire()
                msg = socket.recv(zmq.NOBLOCK)
            except Exception:
                # noblock error, can continue
                self.csuLock.release()
                time.sleep(0.2)
                continue

            self.csuLock.release()
            self.processZmqMesg(msg)

        self.log_debug("Exiting handlePubSubMsgs thread")

    def processZmqMesg(self, msg):
        TYPE_KEEP_ALIVE_REQ = 106
        TYPE_GO_DOWN = 100
        TYPE_DOWN = 101
        TYPE_COME_UP = 102
        TYPE_UP = 103
        TYPE_NETWORK_LAYER_INFO = 109
        TYPE_MODULE_STATUS_REQ = 110

        msgLen = len(msg)

        if (msgLen <= 4):
            self.Log.log_error(
                "The msg received from CSU Manager has invalid mesg length")
            return

        lenInPkt = struct.unpack("i", msg[0:4])[0]
        if msgLen != lenInPkt:
            self.Log.log_error("received mesg length %d, doesn't match length in pkt %d" % (msgLen, lenInPkt))
            return

        index = 4  # offset into first message
        while (index < msgLen):
            bufferType = struct.unpack("i", msg[index:index+4])[0]
            bufferLen = struct.unpack("i", msg[index+4:index+8])[0]
            # buffer = msg[index]
            buffer = self.extractMessage(msg, index, index+bufferLen, msgLen)
            if (bufferType == TYPE_KEEP_ALIVE_REQ):
                self.handleKeepAliveReqMesg(buffer)

            elif (bufferType == TYPE_GO_DOWN):
                self.handleGoDownMesg(buffer)

            elif (bufferType == TYPE_DOWN):
                self.handleDownMesg(buffer)

            elif (bufferType == TYPE_COME_UP):  # Come up
                self.handleComeUpMesg(buffer)

            elif (bufferType == TYPE_UP):
                self.handleUpMesg(buffer)

            elif (bufferType == TYPE_NETWORK_LAYER_INFO):
                self.handleNetworkLayerInfoMesg(buffer)

            elif (bufferType == TYPE_MODULE_STATUS_REQ):
                self.handleModuleStatusReqMesg(buffer)

            else:  # All other messages
                self.log_debug("mesg type %d received" % (bufferType))

            index += bufferLen

    '''
    Once we set PROTODOWN on clag interfaces, wait for switchd to
    bring down the oper status before sending an ack back to CSU Mgr
    '''

    def monitorOperStatus(self):
        self.log_debug("Monitoring oper status")
        for retries in range(0, 10):
            allOperDown = True
            for intf in self.Parser.GetClagIntfDB().values():
                if intf.intfType == self.Parser._CLAG_INTF_TYPE_BOND:
                    intfOperState = self.Intf.GetOperState(intf.intfName)
                    self.log_debug("Intf %s, state %s" %
                            (intf.intfName, intfOperState))

                    if intfOperState == "up":
                        allOperDown = False

            if allOperDown == True or self.keep_going == False:
                break

            time.sleep(0.2)

        self.log_debug("Sending going down reply message")
        self.sendGoDownReplyMsg()

    def setModeAndState(self, mode, state):
        self.mode = mode
        self.state = state

    def setSlavesProtoDown(self, state):
        self.log_debug("setSlavesProtoDown state %d" % (state))
        self.protoDownState = state

        nlm = nlmanager.nlmanager.NetlinkManager(util.CLAG_NLM_CSU_ID)
        self.Intf.protoDownLock.acquire()
        nlmsgs = bytes()
        for intf in self.Parser.GetClagIntfDB().values():
            if intf.intfType == self.Parser._CLAG_INTF_TYPE_BOND:
                self.log_debug("Intf %s, state %s, set protodown %s" %
                        (intf.intfName, self.Intf.GetOperState(intf.intfName), state))
                if state == True:
                    nlmsgs += intf.UpdateProtoDownFlags(nlm, intf._PROTO_DOWNF_CSU,
                                                        0, batch=True)
                else:
                    nlmsgs += intf.UpdateProtoDownFlags(nlm, 0, intf._PROTO_DOWNF_CSU,
                                                        batch=True)

        nlm.tx_nlpacket_raw(nlmsgs)
        self.Intf.protoDownLock.release()

        if state == True:
            self.monitorOperStatusThread = threading.Thread(None, self.monitorOperStatus,
                                                            "Send DownCompl to CS Manager")
            self.monitorOperStatusThread.daemon = True
            self.monitorOperStatusThread.start()

    def isSlaveProtoDown(self):
        return self.protoDownState

