#! /usr/bin/python3
# -------------------------------------------------------------------------------
#
# Copyright 2014, 2020 Cumulus Networks, inc  all rights reserved
#
# -------------------------------------------------------------------------------
#

try:
    import clag.datasync
    import clag.util as util
    import copy
    import json
    import os
    import queue
    import threading
    import nlmanager.nlmanager
    import nlmanager.nlpacket
    import socket
    import struct
    from ipaddress import IPv4Address, IPv6Address
    import xml.etree.ElementTree as ET
    import clag.clagdPeerXchange_pb2 as peerXchange
except ImportError as e:
    raise ImportError(str(e) + "- required module not found")


class mdbsync(clag.datasync.datasync):
    """
    This class handles everything required to synchronize the Multicast 
    Forwarding Database, or MDB, between two CLAG peers.
    """

    # Work queue operation codes.
    _MCAST_ADD = 1
    _MCAST_DELETE = 2
    _MCAST_REFRESH = 3
    _RPORT_ADD = 4
    _RPORT_DELETE = 5
    _RPORT_REFRESH = 6

    # MDB Work Queue operation codes
    _WORK_QUEUE_MCAST_ADD = 1
    _WORK_QUEUE_MCAST_DELETE = 2
    _WORK_QUEUE_RPORT_ADD = 4
    _WORK_QUEUE_RPORT_DELETE = 5
    _WORK_QUEUE_REFRESH = 6
    _WORK_QUEUE_NOP = 7
    _WORK_QUEUE_DUMP_DONE = 8

    # IGMP Fast Leave Notification
    _MDB_FLAGS_FAST_LEAVE = nlmanager.nlpacket.MDB.MDB_FLAGS_FAST_LEAVE

    # Max entries synced to peer in a single protobuf/XML message
    _MAX_SYNC = 500

    def __init__(self, Daemon, Intf, Parser, Log, sendQueue, clagnl):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(mdbsync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Daemon = Daemon
        self.Intf = Intf
        self.Parser = Parser
        self.Log = Log
        self.sendQueue = sendQueue
        self.clagnl = clagnl
        self.peerIf = self.Parser.args.peerIf
        if self.Intf.isSubIfName(self.peerIf):
            self.peerIf = self.Intf.GetMasterOfSubIf(self.peerIf)

        #
        #   Mcast Database - The Multicast groups which are present on dual
        #   connected interfaces (or their sub-interfaces) are tracked in this
        #   database, which is a dictionary.  The key is a tuple containing the
        #   multicast group, bridge, port, and VLAN ID.  The value is a list
        #   containing the dual connected interface, the number of seconds since
        #   until the entry will be removed, and a flag indicating if the
        #   address was added from the peer or learned on this system and a
        #   timestamp when the address was last refreshed.
        #   {
        #      Multicast Group  Bridge  Port    VlanId     dualIf    Age  Remote  Updated
        #       ("234.1.1.2",   "br0",  "bond0", 101) :   [ "bond0",  42, True,       14],
        #       ("234.1.1.3",   "br0",  "bond0", 101) :   [ "bond0",  31, False,       0],
        #       ("234.1.1.4",   "br0",  "bond1", 102) :   [ "bond1", 240, True,       34]
        #   }
        #
        self.mcastDBLock = threading.Lock()
        self.mcastDB = {}

        #   As part of dual link failure handling to retain the entry on the peerlink 
        #   the peerlink is made part of  the multicast groups for which the clag bond
        #   is part of, so even though the host bond goes down, the peerlink entry is retained 
        #   and avoids flooding of multicast traffic. 
        #   The given dictionary mcastPeerlinkUserDB maintains mapping between the 
        #   (multicast group, bridge, peerlink, vlanId) to the users. 
        #   The users here refer the clag bonds. 
        self.mcastPeerlinkUserDBLock = threading.Lock()
        self.mcastPeerlinkUserDB = {}

        #
        #   Router Port Database - The Multicast router ports which are present
        #   on dual connected interfaces (or their sub-interfaces) are tracked
        #   in this database, which is a dictionary.  The key is a tuple
        #   containing the bridge and port.  The value is a list containing the
        #   dual connected interface, the number of seconds since until the
        #   entry will be removed, and a flag indicating if the address was
        #   added from the peer or learned on this system and a timestamp when
        #   the address was last refreshed.
        #   {
        #        Bridge  Port         dualIf   Age  Remote  Updated
        #       ("br0",  "bond0") : [ "bond0",  42, True,       14],
        #       ("br0",  "bond0") : [ "bond0",  31, False,       0],
        #       ("br0",  "bond1") : [ "bond1", 240, True,       34]
        #   }
        #
        # This database is locked with self.mcastDBLock
        self.rportDB = {}

        #
        #   Multicast Work Queue - This queue contains Multicast Groups and
        #   router ports which need to be added, deleted or refreshed.  Each
        #   entry is a tuple containing an operation, the multicast group, the
        #   bridge, the bridge port, and a VLAN Id:
        #
        #       ( _MCAST_ADD, "234.1.1.2", "br0", "bond0", 101 )
        #       ( _RPORT_ADD, None,        "br0", "bond1", 0   )
        #

        # netlink notification received by Kernel are processed by this queue
        self.mdbFromKernWorkQueue = queue.Queue()
        util.queueAddNameMap(self.mdbFromKernWorkQueue, 'mdbFromKernWorkQueue')

        self.mdbToKernWorkQueue = queue.Queue()
        util.queueAddNameMap(self.mdbToKernWorkQueue, 'mdbToKernWorkQueue')

        self.clagnl.SetMdbSync(self)
        self.keep_going = True
        self.stopEvent = clag.clagthread.Event()

        self.updateThread = threading.Thread(
            None, self.UpdateMcastsInHwT, "UpdateMcastsInHw")
        self.updateThread.daemon = True
        self.updateThread.start()

        self.kernelMsgThread = threading.Thread(
            None, self.MdbKernelMsgT, "MdbKernelMsg")
        self.kernelMsgThread.daemon = True
        self.kernelMsgThread.start()

        self.refreshMcastsThread = threading.Thread(
            None, self.RefreshMcastsT, "RefreshMcasts")
        self.refreshMcastsThread.daemon = True
        self.refreshMcastsThread.start()

    def __del__(self):
        """
        Destructor:
        """
        #
        #   Signal the update thread to wake up and exit
        #
        self.SignalThreads()
        self.JoinThreads()

    def RefreshMcastsT(self):
        try:
            self.RefreshMcasts()
        except:
            self.Daemon.DumpTraceback()

    def MdbKernelMsgT(self):
        try:
            self.MdbKernelMsg()
        except:
            self.Daemon.DumpTraceback()

    def UpdateMcastsInHwT(self):
        try:
            self.Log.log("[%d]Init UpdateMcastsInHw" % (self.Parser.getTid()))
            self.UpdateMcastsInHw()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.stopEvent.set()
        try:
            self.mdbToKernWorkQueue.put_nowait((None, None, None, None, None))
            util.queueSetMaxLen('mdbToKernWorkQueue')
        except queue.Full:
            pass
        try:
            self.mdbFromKernWorkQueue.put_nowait((mdbsync._WORK_QUEUE_NOP, ()))
            util.queueSetMaxLen('mdbFromKernWorkQueue')
        except queue.Full:
            pass

    def JoinThreads(self):
        self.updateThread.join()
        self.kernelMsgThread.join()
        self.refreshMcastsThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_MDB_SYNC, logStr)

    def log_verbose(self, logStr):
        self.Log.log_verbose(self.Log._LOG_DEBUG_MDB_SYNC, logStr)

    def ToKernWorkQueuePut(self, element):
        if self.keep_going:
            self.mdbToKernWorkQueue.put(element)
            util.queueSetMaxLen('mdbToKernWorkQueue')

    def GetObjectSize(self):
        mcastObj = {}
        self.mcastDBLock.acquire()
        mcastObj['mcastDB'] = util.getSize(self.mcastDB)
        mcastObj['rportDB'] = util.getSize(self.rportDB)
        self.mcastDBLock.release()
        return mcastObj

    def __str__(self):
        """
        Return local information in human-readable form
        """
        mcastDb = []
        self.mcastDBLock.acquire()
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in list(self.mcastDB.items()):
            if not remote:
                mcastInfo = {}
                mcastInfo['vlanId'] = vlanId
                mcastInfo['group'] = group
                mcastInfo['bridge'] = bridge
                mcastInfo['port'] = port
                mcastInfo['age'] = age
                mcastDb.append(mcastInfo)
        self.mcastDBLock.release()
        return json.dumps(mcastDb, sort_keys=True, indent=4)

    def EncodedOurStr(self):
        return self.__str__()

    def EncodedPeerStr(self):
        """
        Return peer information in human-readable form
        """
        mcastDb = []
        self.mcastDBLock.acquire()
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in list(self.mcastDB.items()):
            if remote:
                mcastInfo = {}
                mcastInfo['vlanId'] = vlanId
                mcastInfo['group'] = group
                mcastInfo['bridge'] = bridge
                mcastInfo['port'] = port
                mcastInfo['age'] = age
                mcastDb.append(mcastInfo)
        self.mcastDBLock.release()
        return json.dumps(mcastDb, sort_keys=True, indent=4)

    def EncodedOurRportStr(self):
        """
        Return local router port information in human-readable form
        """
        mcastDb = []
        self.mcastDBLock.acquire()
        for (bridge, port), [dualIf, age, remote, updated] in list(self.rportDB.items()):
            if not remote:
                mcastInfo = {}
                mcastInfo['bridge'] = bridge
                mcastInfo['port'] = port
                mcastInfo['dualIf'] = dualIf
                mcastInfo['age'] = age
                mcastDb.append(mcastInfo)
        self.mcastDBLock.release()
        return json.dumps(mcastDb, sort_keys=True, indent=4)

    def EncodedPeerRportStr(self):
        """
        Return local router port information in human-readable form
        """
        mcastDb = []
        self.mcastDBLock.acquire()
        for (bridge, port), [dualIf, age, remote, updated] in list(self.rportDB.items()):
            if remote:
                mcastInfo = {}
                mcastInfo['bridge'] = bridge
                mcastInfo['port'] = port
                mcastInfo['dualIf'] = dualIf
                mcastInfo['age'] = age
                mcastDb.append(mcastInfo)
        self.mcastDBLock.release()
        return json.dumps(mcastDb, sort_keys=True, indent=4)

    def CollectLocalInfo(self, nlm):
        """
        This function is called on a periodic interval to gather up any 
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.
        """
        pass

    def GetBridgeMembershipIntervals(self, intfs):
        membershipIntervals = {}
        masterVlanMap = self.Intf.GetMasterIfVlanMap()
        for intf in intfs:
            age = self.Intf.GetBridgeMemInterval(
                self.Intf.GetBridgeOfMember(intf))
            if age is not None:
                membershipIntervals[intf] = age
            for subIf in list(masterVlanMap.get(intf, {}).values()):
                age = self.Intf.GetBridgeMemInterval(
                    self.Intf.GetBridgeOfMember(subIf))
                if age is not None:
                    membershipIntervals[subIf] = age
        return membershipIntervals

    def GetBridgeQuerierIntervals(self, intfs):
        querierIntervals = {}
        masterVlanMap = self.Intf.GetMasterIfVlanMap()
        for intf in intfs:
            age = self.Intf.GetBridgeQuerierInterval(
                self.Intf.GetBridgeOfMember(intf))
            if age is not None:
                querierIntervals[intf] = age
            for subIf in list(masterVlanMap.get(intf, {}).values()):
                age = self.Intf.GetBridgeQuerierInterval(
                    self.Intf.GetBridgeOfMember(subIf))
                if age is not None:
                    querierIntervals[subIf] = age
        return querierIntervals

    def AddPortToMulitcastGroup(self, bridge, kvid, group, port, user):
        if port is None or user is None:
            return 

        mcastKey = (group, bridge, port, kvid)
        if not self.Intf.isVlanFilteringBridge(bridge):
            kvid = 0
        if mcastKey not in self.mcastDB: 
            self.mcastDB[mcastKey] = [port, 0, False, 0]
            self.ToKernWorkQueuePut((mdbsync._MCAST_ADD, group, bridge, port, kvid))
            self.log_debug("Added peerlink %s to Multicast group %s bridge %s vlanId %s" %
                             (port, group, bridge, kvid))
        else:
            self.ToKernWorkQueuePut((mdbsync._MCAST_REFRESH, group, bridge, port, kvid)) 

        self.mcastPeerlinkUserDBLock.acquire()
        if mcastKey not in self.mcastPeerlinkUserDB:
            self.mcastPeerlinkUserDB[mcastKey] = [user]
        else:
            if user not in self.mcastPeerlinkUserDB[mcastKey]:
                self.mcastPeerlinkUserDB[mcastKey].append(user)
        self.mcastPeerlinkUserDBLock.release()

    def DelPeerlinkMulticastUser(self, bridge, kvid, group, port, user):
        if port is None or user is None:
            return
        
        mcastKey = (group, bridge, port, kvid)
        if not self.Intf.isVlanFilteringBridge(bridge):
            kvid = 0 

        if mcastKey in self.mcastPeerlinkUserDB:
            self.mcastPeerlinkUserDBLock.acquire()
            if user in self.mcastPeerlinkUserDB[mcastKey]:
                self.mcastPeerlinkUserDB[mcastKey].remove(user)
            if not self.mcastPeerlinkUserDB[mcastKey]:
                if mcastKey in self.mcastDB:
                    del self.mcastDB[mcastKey]
                del self.mcastPeerlinkUserDB[mcastKey]
                self.ToKernWorkQueuePut((mdbsync._MCAST_DELETE, group, bridge, port, kvid))
                self.log_debug("IGMP Fast Leave Received. Deleted peerlink %s to Multicast group %s bridge %s vlanId %s" %
                            (port, group, bridge, kvid))
            self.mcastPeerlinkUserDBLock.release()

    def AddMcastXmlEntry(self, element, op, dualIf, bridge, age, vlanId, group):
        """
        Adds an Mcast Xml entry to the root element
        """

        mcast = ET.SubElement(element, "multicast")
        mcastAttrib = {"op": op, "dualIf": str(
            dualIf), "dev": str(bridge), "age": str(age)}
        if vlanId is not None:
            mcastAttrib["vlan"] = str(vlanId)
        ET.SubElement(mcast, "group", mcastAttrib).text = group

    def AddRouterXmlEntry(self, element, op, dualIf, bridge, age):
        """
        Adds a Router Xml entry to the root element
        """

        mcast = ET.SubElement(element, "multicast")
        rportAttrib = {"op": op, "dualIf": str(
            dualIf), "dev": str(bridge), "age": str(age)}
        ET.SubElement(mcast, "router_port", rportAttrib)

    def AddProtoBufEntry(self, data, entryType, op, dualIf, bridge, age, vlanId=None, group=None):
        """
        Adds an Mcast Xml entry to the root element
        """

        entry = data.entry.add()
        entry.type = entryType
        entry.op = op
        entry.dualIf = str(dualIf)
        entry.brigge = str(bridge)
        entry.age = str(age)
        if vlanId:
            entry.vlanId = vlanId
        if group:
            entry.group = group

    def QueueMessage(self, protobuf, msgData):
        if protobuf:
            return self.QueueProtoBufMessage(msgData)
        else:
            return self.QueueXmlMessage(msgData)

    def QueueXmlMessage(self, msgData):
        xmlStr = ET.tostring(msgData)
        self.sendQueue.put((None, xmlStr))
        msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
        return msgData

    def QueueProtoBufMessage(self, msgData):
        if not msgData.ByteSize():
            return peerXchange.MdbSync()
        self.Log.log_debug(self.Log._LOG_DEBUG_INFORM_PEER, "TXing mdb data: %s" % str(msgData))
        msgHdr = peerXchange.Header()
        msgHdr.type = peerXchange.Header.MessageType.Value('CLAG_MDB_SYNC')
        msgHdr.data = msgData.SerializeToString()
        self.sendQueue.put((msgHdr, None))
        msgData = peerXchange.MdbSync()
        return msgData

    def MakeProtoBufMessage(self):
        """
        Writes the local information as a protocol buffer message.
        """
        if self.Parser.syncDoneToPeer:
            return

        msgData = peerXchange.MdbSync()
        self.mcastDBLock.acquire()
        c_mcastDB = copy.deepcopy(self.mcastDB)
        c_rportDB = copy.deepcopy(self.rportDB)
        self.mcastDBLock.release()
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in c_mcastDB.items():
            if not remote:
                self.AddProtoBufEntry(msgData, "group", "mdb-sync",
                                      dualIf, bridge, age, vlanId, group)
        for (bridge, port), [dualIf, age, remote, update] in c_rportDB.items():
            if not remote:
                self.AddProtoBufEntry(msgData, "router", "mdb-sync", dualIf, bridge, age)
        self.QueueProtoBufMessage(msgData)

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML.

        parent - An xml.etree.ElementTree.Element instance which is the parent
            element under which all of the information should be encoded.

        <multicast>
            <group dualIf="bond2" vlan="100" dev="br0" age="259">232.10.10.10</group>
            <group dualIf="bond3" vlan="101" dev="br1" age="13">232.10.10.11</group>
            <group dualIf="bond4" dev="br2" age="125">232.10.10.12</group>
            <router_port dualIf="bond4" dev="br2" age="125"/>
            <router_port dualIf="bond0" dev="br0" age="14"/>
        </multicast>
        """
        if self.Parser.syncDoneToPeer:
            return

        self.mcastDBLock.acquire()
        c_mcastDB = copy.deepcopy(self.mcastDB)
        c_rportDB = copy.deepcopy(self.rportDB)
        self.mcastDBLock.release()
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in list(c_mcastDB.items()):
            if not remote:
                self.AddMcastXmlEntry(parent, "mdb-sync",
                                      dualIf, bridge, age, vlanId, group)

        for (bridge, port), [dualIf, age, remote, update] in list(c_rportDB.items()):
            if not remote:
                self.AddRouterXmlEntry(parent, "mdb-sync", dualIf, bridge, age)

    def GetWorkQueueDelayed(self, delay=0.001):
        '''
        Returns the next entry on the MDB work queue, waiting up to the specified
        delay (default 1 millisecond) for an entry.
        '''
        queueEmpty = False
        try:
            (opcode, k_params) = self.mdbFromKernWorkQueue.get(True, delay)
        except queue.Empty:
            opcode = mdbsync._WORK_QUEUE_NOP
            k_params = {}
            queueEmpty = True
        return (opcode, k_params, queueEmpty)

    def SyncCurrentStateToPeer(self, kMdb):
        """
        Syncs to Local Mulitcast Entry to the peer
        """

        self.log_verbose("Beginning sync of local addresses")
        self.mcastDBLock.acquire()
        mcastDB = copy.deepcopy(self.mcastDB)
        rportDB = copy.deepcopy(self.rportDB)
        self.mcastDBLock.release()
        kmcastDB = kMdb["mdba-mdb"]
        krportDB = kMdb["mdba-router"]
        Entries = 0
        peerIfs = set(self.Intf.GetSubIfsOfMaster(
            self.peerIf)) | set([self.peerIf])
        protobuf = self.Parser.peerProtoBuf
        if protobuf:
            msgData = peerXchange.MdbSync()
        else:
            msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})

        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in list(mcastDB.items()):
            isPeerLink = True if port in peerIfs else False
            mcastEntryState = kmcastDB.get(
                (group, bridge, port, vlanId), (None, None, None))[1]
            if not remote and not isPeerLink and mcastEntryState != 1:
                if protobuf:
                    self.AddProtoBufEntry(msgData, "group", "lazy-sync", dualIf, bridge, age, vlanId, group)
                else:
                    self.AddMcastXmlEntry(msgData, "lazy-sync", dualIf, bridge, age, vlanId, group)
                Entries += 1
                if Entries >= mdbsync._MAX_SYNC and self.keep_going:
                    msgData = self.QueueMessage(protobuf, msgData)
                    Entries = 0

        for (bridge, port), [dualIf, age, remote, updated] in list(rportDB.items()):
            rportState = self.Intf.getBridgeMcastRtrType(port)
            # Do not sync permanent router port entries
            if not remote and rportState != nlmanager.nlpacket.MDB.MDB_RTR_TYPE_PERM:
                if protobuf:
                    self.AddProtoBufEntry(msgData, "router", "lazy-sync", dualIf, bridge, age)
                else:
                    self.AddRouterXmlEntry(msgData, "lazy-sync", dualIf, bridge, age)
                Entries += 1
                if Entries >= mdbsync._MAX_SYNC and self.keep_going:
                    msgData = self.QueueMessage(protobuf, msgData)
                    Entries = 0
        if Entries and self.keep_going:
            self.QueueMessage(protobuf, msgData)
        self.log_verbose("Ending sync of local addresses")

    def ParseProtoBufMessage(self, msgHdr):
        """
        Reads the protobuf information sent from the peer.
        """
        self.log_verbose("Beginning ParseProtoBufMessage")
        msgData = peerXchange.MdbSync()
        msgData.ParseFromString(msgHdr.data)
        # Get the current list of dual connected hosts
        dualIfs = self.Intf.GetDualIfs()
        revDualIfs = {v: k for k, v in list(dualIfs.items())}
        # Get a mapping of master interfaces to sub-interface names and VLAN IDs
        vlanMap = self.Intf.GetMasterIfVlanMap()
        # For each bridge with dual connected interfaces, get the multicast membership interval.
        bridgeMemIntval = None
        bridgeQuerierIntval = None

        peerlinkOfBridge = self.Intf.GetPeerlinksOfBridgeMap()
        peerlinkMdbRefreshTracker = []

        self.mcastDBLock.acquire()
        now = int(os.times()[4])
        for entry in msgData.entry:
            op = entry.op
            r_type = entry.type
            r_peerDualIf = entry.dualIf
            r_dualIf = revDualIfs.get(r_peerDualIf)
            r_age = int(entry.age)

            if r_type == "group":
                r_group = entry.group
                r_vlanId = entry.vlanId

                # Convert the r_dualIf, r_vlanId into a bridge, port, vlanId
                port = vlanMap.get(r_dualIf, {}).get(r_vlanId, r_dualIf)
                bridge = self.Intf.GetBridgeOfMember(port)
                self.log_debug("Multicast group Receive From Peer %s bridge %s myport %s \
                myDualIf %s remote dualIf %s vlanId %s age %s" %
                               (r_group, bridge, port, r_dualIf, r_peerDualIf, r_vlanId, r_age))

                if bridge is not None:
                    mcastKey = (r_group, bridge, port, r_vlanId)
                    if not self.Intf.isVlanFilteringBridge(bridge):
                        kvid = 0
                    else:
                        kvid = r_vlanId
                    if op in ("mcast-add", "lazy-sync"):
                        if mcastKey not in self.mcastDB:
                            self.mcastDB[mcastKey] = [
                                r_dualIf, r_age, True, now]
                            self.ToKernWorkQueuePut(
                                (mdbsync._MCAST_ADD, r_group, bridge, port, kvid))
                            self.log_debug("Added new remote: Multicast group %s bridge %s port %s vlanId %s age %s" %
                                           (r_group, bridge, port, r_vlanId, r_age))
                        else:
                            [c_dualIf, c_age, c_remote,
                                c_updated] = self.mcastDB[mcastKey]
                            if c_remote:
                                if bridgeMemIntval is None:
                                    bridgeMemIntval = self.GetBridgeMembershipIntervals(
                                        dualIfs)

                                max_age = bridgeMemIntval.get(port, 260)

                                c_updated = now
                                self.ToKernWorkQueuePut(
                                    (mdbsync._MCAST_REFRESH, r_group, bridge, port, kvid))
                                self.log_debug("Refreshing remote: Multicast group %s bridge %s port %s vlanId %s age %s" %
                                               (r_group, bridge, port, r_vlanId, r_age))
                                self.mcastDB[mcastKey] = [
                                    r_dualIf, r_age, True, c_updated]
                            else:
                                if r_age > c_age:
                                    self.mcastDB[mcastKey] = [
                                        r_dualIf, r_age, True, now]
                                    self.ToKernWorkQueuePut(
                                        (mdbsync._MCAST_REFRESH, r_group, bridge, port, kvid))
                                    self.log_debug("Moving local to remote: Multicast group %s bridge %s port %s vlanId %s age %s" %
                                                   (r_group, bridge, port, r_vlanId, r_age))
                        if (bridge, r_vlanId, r_group) not in peerlinkMdbRefreshTracker:
                            peerlinkMdbRefreshTracker.append((bridge, r_vlanId, r_group))
                            peerlinkIntf = peerlinkOfBridge.get(bridge, None)
                            self.AddPortToMulitcastGroup(bridge, r_vlanId, r_group, peerlinkIntf, port)
                    else:
                        if mcastKey in self.mcastDB:
                            self.log_debug(
                                "Remote Delete Notification. Igmp Fast Leave rxed on peer")
                            del self.mcastDB[mcastKey]
                            self.ToKernWorkQueuePut(
                                (mdbsync._MCAST_DELETE, r_group, bridge, port, kvid))
                            peerlinkIntf = peerlinkOfBridge.get(bridge, None)
                            self.DelPeerlinkMulticastUser(bridge, r_vlanId, r_group, peerlinkIntf, port)
                            self.log_debug(" Entry deleted: Multicast group %s bridge %s port %s vlanId %s age %s" %
                                           (r_group, bridge, port, r_vlanId, r_age))

            if r_type == "router":
                # Convert the r_dualIf into a bridge
                bridge = self.Intf.GetBridgeOfMember(r_dualIf)
                port = r_dualIf
                if bridge is not None:
                    rportKey = (bridge, port)
                    if op in ("rport-add", "lazy-sync"):
                        if rportKey not in self.rportDB:
                            self.rportDB[rportKey] = [
                                r_dualIf, r_age, True, now]
                            self.ToKernWorkQueuePut(
                                (mdbsync._RPORT_ADD, None, bridge, port, 0))
                            self.log_debug("Added new remote: Multicast router port %s bridge %s age %s" % (
                                port, bridge, r_age))
                        else:
                            [c_dualIf, c_age, c_remote,
                                c_updated] = self.rportDB[rportKey]
                            if c_remote:
                                if bridgeQuerierIntval is None:
                                    bridgeQuerierIntval = self.GetBridgeQuerierIntervals(
                                        dualIfs)
                                max_age = bridgeQuerierIntval.get(port, 255)
                                if (now - c_updated) > (max_age * 2/3):
                                    c_updated = now
                                    self.ToKernWorkQueuePut(
                                        (mdbsync._RPORT_REFRESH, None, bridge, port, 0))
                                    self.log_debug("Refreshing remote: Multicast router port %s bridge %s age %s" % (
                                        port, bridge, r_age))

                                self.rportDB[rportKey] = [
                                    r_dualIf, r_age, True, c_updated]
                            else:
                                if r_age > c_age:
                                    self.rportDB[rportKey] = [
                                        r_dualIf, r_age, True, now]
                                    self.ToKernWorkQueuePut(
                                        (mdbsync._RPORT_REFRESH, None, bridge, port, 0))
                                    self.log_debug("Moving local to remote: Multicast router port %s bridge %s age %s" % (
                                        port, bridge, r_age))
                    else:
                        if rportKey in self.rportDB:
                            #del self.rportDB[rportKey]
                            #self.ToKernWorkQueuePut((mdbsync._RPORT_DELETE, None, bridge, port, 0))
                            self.log_debug("Multicast router Entry Delete port %s bridge %s age %s" % (
                                port, bridge, r_age))

        self.mcastDBLock.release()
        self.log_verbose("Ending ParseProtoBufMessage")
        return msgData

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.
        root - The root element of the XML tree
        """
        if root.find('multicast') is not None:
            # Get the current list of dual connected hosts
            dualIfs = self.Intf.GetDualIfs()
            revDualIfs = {v: k for k, v in list(dualIfs.items())}
            peerlinkOfBridge = self.Intf.GetPeerlinksOfBridgeMap()
            peerlinkMdbRefreshTracker = [] 

            # Get a mapping of master interfaces to sub-interface names and VLAN IDs
            vlanMap = self.Intf.GetMasterIfVlanMap()

        # For each bridge with dual connected interfaces, get the multicast membership interval.
        bridgeMemIntval = None
        bridgeQuerierIntval = None

        self.mcastDBLock.acquire()
        now = int(os.times()[4])
        for mcast in root.findall('multicast'):
            for group_element in mcast.findall('group'):
                op = group_element.get('op')
                r_peerDualIf = group_element.get('dualIf')
                r_dualIf = revDualIfs.get(r_peerDualIf)
                r_age = int(group_element.get('age'))
                r_group = group_element.text
                r_vlanId = group_element.get('vlan')

                if r_vlanId is not None:
                    r_vlanId = int(r_vlanId)

                # Convert the r_dualIf, r_vlanId into a bridge, port, vlanId
                port = vlanMap.get(r_dualIf, {}).get(r_vlanId, r_dualIf)
                bridge = self.Intf.GetBridgeOfMember(port)
                self.log_debug("Multicast group Receive From Peer %s bridge %s myport %s \
                myDualIf %s remote dualIf %s vlanId %s age %s" %
                               (r_group, bridge, port, r_dualIf, r_peerDualIf, r_vlanId, r_age))

                if bridge is not None:
                    mcastKey = (r_group, bridge, port, r_vlanId)
                    if not self.Intf.isVlanFilteringBridge(bridge):
                        kvid = 0
                    else:
                        kvid = r_vlanId
                    if op in ("mcast-add", "lazy-sync"):
                        if mcastKey not in self.mcastDB:
                            self.mcastDB[mcastKey] = [
                                r_dualIf, r_age, True, now]
                            self.ToKernWorkQueuePut(
                                (mdbsync._MCAST_ADD, r_group, bridge, port, kvid))
                            self.log_debug("Added new remote: Multicast group %s bridge %s port %s vlanId %s age %s" %
                                           (r_group, bridge, port, r_vlanId, r_age))
                        else:
                            [c_dualIf, c_age, c_remote,
                                c_updated] = self.mcastDB[mcastKey]
                            if c_remote:
                                if bridgeMemIntval is None:
                                    bridgeMemIntval = self.GetBridgeMembershipIntervals(
                                        dualIfs)

                                max_age = bridgeMemIntval.get(port, 260)

                                c_updated = now
                                self.ToKernWorkQueuePut(
                                    (mdbsync._MCAST_REFRESH, r_group, bridge, port, kvid))
                                self.log_debug("Refreshing remote: Multicast group %s bridge %s port %s vlanId %s age %s" %
                                               (r_group, bridge, port, r_vlanId, r_age))
                                self.mcastDB[mcastKey] = [
                                    r_dualIf, r_age, True, c_updated]
                            else:
                                if r_age > c_age:
                                    self.mcastDB[mcastKey] = [
                                        r_dualIf, r_age, True, now]
                                    self.ToKernWorkQueuePut(
                                        (mdbsync._MCAST_REFRESH, r_group, bridge, port, kvid))
                                    self.log_debug("Moving local to remote: Multicast group %s bridge %s port %s vlanId %s age %s" %
                                                   (r_group, bridge, port, r_vlanId, r_age))
                        if (bridge, r_vlanId, r_group) not in peerlinkMdbRefreshTracker:
                            peerlinkMdbRefreshTracker.append((bridge, r_vlanId, r_group))
                            peerlinkIntf = peerlinkOfBridge.get(bridge, None)
                            self.AddPortToMulitcastGroup(bridge, r_vlanId, r_group, peerlinkIntf, port)
                    else:
                        if mcastKey in self.mcastDB:
                            self.log_debug(
                                "Remote Delete Notification. Igmp Fast Leave rxed on peer")
                            del self.mcastDB[mcastKey]
                            self.ToKernWorkQueuePut(
                                (mdbsync._MCAST_DELETE, r_group, bridge, port, kvid))
                            peerlinkIntf = peerlinkOfBridge.get(bridge, None)
                            self.DelPeerlinkMulticastUser(bridge, r_vlanId, r_group, peerlinkIntf, port)
                            self.log_debug(" Entry deleted: Multicast group %s bridge %s port %s vlanId %s age %s" %
                                           (r_group, bridge, port, r_vlanId, r_age))

            for rport_element in mcast.findall('router_port'):
                op = rport_element.get('op')
                r_peerDualIf = rport_element.get('dualIf')
                r_dualIf = revDualIfs.get(r_peerDualIf)
                r_age = int(rport_element.get('age'))

                # Convert the r_dualIf into a bridge
                bridge = self.Intf.GetBridgeOfMember(r_dualIf)
                port = r_dualIf
                if bridge is not None:
                    rportKey = (bridge, port)
                    if op in ("rport-add", "lazy-sync"):
                        if rportKey not in self.rportDB:
                            self.rportDB[rportKey] = [
                                r_dualIf, r_age, True, now]
                            self.ToKernWorkQueuePut(
                                (mdbsync._RPORT_ADD, None, bridge, port, 0))
                            self.log_debug("Added new remote: Multicast router port %s bridge %s age %s" % (
                                port, bridge, r_age))
                        else:
                            [c_dualIf, c_age, c_remote,
                                c_updated] = self.rportDB[rportKey]
                            if c_remote:
                                if bridgeQuerierIntval is None:
                                    bridgeQuerierIntval = self.GetBridgeQuerierIntervals(
                                        dualIfs)
                                max_age = bridgeQuerierIntval.get(port, 255)
                                if (now - c_updated) > (max_age * 2/3):
                                    c_updated = now
                                    self.ToKernWorkQueuePut(
                                        (mdbsync._RPORT_REFRESH, None, bridge, port, 0))
                                    self.log_debug("Refreshing remote: Multicast router port %s bridge %s age %s" % (
                                        port, bridge, r_age))

                                self.rportDB[rportKey] = [
                                    r_dualIf, r_age, True, c_updated]
                            else:
                                if r_age > c_age:
                                    self.rportDB[rportKey] = [
                                        r_dualIf, r_age, True, now]
                                    self.ToKernWorkQueuePut(
                                        (mdbsync._RPORT_REFRESH, None, bridge, port, 0))
                                    self.log_debug("Moving local to remote: Multicast router port %s bridge %s age %s" % (
                                        port, bridge, r_age))
                    else:
                        if rportKey in self.rportDB:
                            #del self.rportDB[rportKey]
                            #self.ToKernWorkQueuePut((mdbsync._RPORT_DELETE, None, bridge, port, 0))
                            self.log_debug("Multicast router Entry Delete port %s bridge %s age %s" % (
                                port, bridge, r_age))

        self.mcastDBLock.release()

    def DualIfChange(self, dualIf, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        #Sync mdb entries to peer
        Entries = 0
        protobuf = self.Parser.peerProtoBuf
        if protobuf:
            msgData = peerXchange.MdbSync()
        else:
            msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
        operState = {}
        for s_if in dualIf:
            operState[s_if] = self.Intf.GetOperStateFromSysFs(s_if)
        self.mcastDBLock.acquire()
        for (group, bridge, port, vlanId), [_dualIf, age, remote, updated] in list(self.mcastDB.items()):
            if _dualIf in dualIf:
                if remote and not enable and operState[_dualIf] == "up":
                    #If entry is remote and dualbond became singly connected
                    #take ownership of entry if operstate is up
                    self.mcastDB[(group, bridge, port, vlanId)] = [_dualIf, age, False, updated]
                elif not remote and enable:
                    #If entry is local send to peer when hostbond becomes dual connected
                    if protobuf:
                        self.AddProtoBufEntry(msgData, "group", "mcast-add", _dualIf, bridge, age, vlanId, group)
                    else:
                        self.AddMcastXmlEntry(msgData, "mcast-add", _dualIf, bridge, age, vlanId, group)
                    Entries += 1
                    if Entries >= mdbsync._MAX_SYNC and self.keep_going:
                        msgData = self.QueueMessage(protobuf, msgData)
                        Entries = 0
        self.mcastDBLock.release()

        if Entries and self.keep_going:
            self.QueueMessage(protobuf, msgData)

    def MatchIdIfChange(self, dualIf, enable):
        """
        Called when an interface gets or loses a matching clag-id on the peer
        """
        pass

    def ClearPeerInfo(self):
        """
        We don't clear multicast group information when the peer link goes down
        or when we exit clagd. So this is a NOP.
        """
        pass

    def ClearLocalInfo(self):
        """
        Removes multicast groups which are owned by this switch.
        """
        self.mcastDBLock.acquire()
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in list(self.mcastDB.items()):
            if not remote:
                self.ToKernWorkQueuePut(
                    (mdbsync._MCAST_DELETE, group, bridge, port, vlanId))
                del self.mcastDB[(group, bridge, port, vlanId)]
        for (bridge, port), [dualIf, age, remote, updated] in list(self.rportDB.items()):
            if not remote:
                self.ToKernWorkQueuePut(
                    (mdbsync._RPORT_DELETE, None, bridge, port, 0))
                del self.rportDB[(bridge, port)]
        self.mcastDBLock.release()

    def GetKernelMdb(self, nlm):
        """
        Retrieve the kernel's multicast database.
        Database format
        kmcastDB = {"mdba-mdb": {(k_group, k_bridge, k_port, k_vid):(k_age, k_state, k_flags), .. }
        """
        kMdb = {}
        kMdb["mdba-mdb"] = {}
        kMdb["mdba-router"] = {}

        debug = nlmanager.nlpacket.RTM_GETMDB in nlm.debug
        mdbMsgList = nlm.request_dump(
            nlmanager.nlpacket.RTM_GETMDB, socket.AF_BRIDGE, debug)
        for msg in mdbMsgList:
            k_bridge = self.Intf.GetIfName(msg.ifindex)
            mdbaMdbList = msg.get_attribute_value(msg.MDBA_MDB)
            mdbaRouterList = msg.get_attribute_value(msg.MDBA_ROUTER)
            if mdbaMdbList:
                for mdbaMdbDict in mdbaMdbList:
                    mdbaMdbList = mdbaMdbDict.get(msg.MDBA_MDB_ENTRY)
                    if mdbaMdbList is not None:
                        for mdbaMdbEntryInfoDict in mdbaMdbList:
                            mdbaMdbEntry = mdbaMdbEntryInfoDict.get(
                                msg.MDBA_MDB_ENTRY_INFO)
                            if mdbaMdbEntry:
                                mdbaMdbEntry[0] = self.Intf.GetIfName(
                                    mdbaMdbEntry[0])
                                mdbaMdbEntry[4] = str(mdbaMdbEntry[4])
                                (k_port, k_state, k_flags, k_vid,
                                 k_group, k_timer_dict) = mdbaMdbEntry
                                k_age = int(
                                    k_timer_dict[msg.MDBA_MDB_EATTR_TIMER])
                                kMdb["mdba-mdb"][(k_group, k_bridge, k_port,
                                                  k_vid)] = (k_age, k_state, k_flags)
            if mdbaRouterList:
                for mdbaRouterDict in mdbaRouterList:
                    mdbaRouterEntry = mdbaRouterDict.get(msg.MDBA_ROUTER_PORT)
                    if mdbaRouterEntry is not None:
                        mdbaRouterEntry[0] = self.Intf.GetIfName(
                            mdbaRouterEntry[0])
                        (k_port, k_timer_dict, k_type_dict) = mdbaRouterEntry
                        k_age_str = k_timer_dict.get(
                            msg.MDBA_ROUTER_PATTR_TIMER)
                        if k_age_str is not None:
                            k_age = int(k_age_str)
                        k_type = k_type_dict.get(msg.MDBA_ROUTER_PATTR_TYPE)
                        kMdb["mdba-router"][(k_bridge, k_port)
                                            ] = (k_age, k_type)

        self.log_debug("Kernel MDB = %s" % str(kMdb))
        return kMdb

    def DoRefreshMcasts(self, nlm):
        # Get the current list of dual connected interfaces

        dualIfs = self.Intf.GetDualIfs()

        # Get a mapping of VLAN sub-interfaces to their master interfaces and VLAN IDs
        vlanMap = self.Intf.GetSubIfsVlanMap()
        peerIfs = set(self.Intf.GetSubIfsOfMaster(
            self.peerIf)) | set([self.peerIf])
        
        peerlinkMdbRefreshTracker = []
        peerlinkOfBridge = self.Intf.GetPeerlinksOfBridgeMap()

        # For each bridge with dual connected interfaces, get the multicast membership
        # interval and querier interval.
        bridgeMemIntval = None
        bridgeQuerierIntval = None

        kMdb = self.GetKernelMdb(nlm)

        self.mcastDBLock.acquire()
        c_mcastDB = copy.deepcopy(self.mcastDB)
        c_rportDB = copy.deepcopy(self.rportDB)
        self.mcastDBLock.release()

        now = int(os.times()[4])

        if kMdb.get("mdba-mdb") is not None:
            for (k_group, k_bridge, k_port, k_vid), (k_age, k_state, k_flags) in list(kMdb["mdba-mdb"].items()):
                if not k_vid:
                    k_vid = vlanMap.get(k_port, (k_port, 0))[1]
                [c_dualIf, c_age, c_remote, c_updated] = c_mcastDB.get(
                    (k_group, k_bridge, k_port, k_vid), [None, None, None, None])
                mcastKey = (k_group, k_bridge, k_port, k_vid)
                # Is the multicast group owned by the peer?
                if c_remote:
                    if k_port not in peerIfs:
                        # Have we seen the address more recently than the peer? 
                        if bridgeMemIntval is None:
                            bridgeMemIntval = self.GetBridgeMembershipIntervals(
                                dualIfs)

                        max_age = bridgeMemIntval.get(k_port, 260)
                        if (k_age - 15) >= (max_age - (now - c_updated)):
                            # Yes. Make the address local
                            self.mcastDBLock.acquire()
                            self.mcastDB[mcastKey] = [
                                c_dualIf, k_age, False, 0]
                            self.mcastDBLock.release()
                            self.log_debug("Moving to local: multicast group %s bridge %s port %s \
                            vlanId %s dualIf %s age %s" %
                                           (k_group, k_bridge, k_port, k_vid, c_dualIf, k_age))

                else:
                    # Multicast group is owned by this switch. Update/add Mcast entry
                    self.mcastDBLock.acquire()
                    if mcastKey not in self.mcastDB:
                        self.mdbFromKernWorkQueue.put((mdbsync._WORK_QUEUE_MCAST_ADD,
                                                       [k_bridge, k_port, k_state, k_flags, k_vid, k_group]))
                    else:
                        self.mcastDB[mcastKey] = [c_dualIf, k_age, False, 0]
                        if k_port not in peerIfs:
                            if (k_bridge, k_vid, k_group) not in peerlinkMdbRefreshTracker:
                                peerlinkMdbRefreshTracker.append((k_bridge, k_vid, k_group))
                                peerlinkIntf = peerlinkOfBridge.get(k_bridge, None)
                                self.AddPortToMulitcastGroup(k_bridge, k_vid, k_group, peerlinkIntf, k_port)
                    self.mcastDBLock.release()
                try:
                    c_mcastDB.pop((k_group, k_bridge, k_port, k_vid), None)
                except KeyError:
                    self.log_debug(" No Entry")

        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in list(c_mcastDB.items()):
            self.mdbFromKernWorkQueue.put((mdbsync._WORK_QUEUE_MCAST_DELETE, [
                                          bridge, port, 0, 0, vlanId, group]))

        if kMdb.get("mdba-router") is not None:
            for (k_bridge, k_port), (k_age, k_type) in list(kMdb["mdba-router"].items()):
                [c_dualIf, c_age, c_remote, c_updated] = c_rportDB.get(
                    (k_bridge, k_port), [None, None, None, None])
                rportKey = (k_bridge, k_port)

                if c_remote:
                    # Have we seen the router port more recently than the peer?
                    if bridgeQuerierIntval is None:
                        bridgeQuerierIntval = self.GetBridgeQuerierIntervals(
                            dualIfs)

                    max_age = bridgeQuerierIntval.get(k_port, 255)
                    if (k_age - 15) >= (max_age - (now - c_updated)):
                        # Yes. Make the router port local
                        self.mcastDBLock.acquire()
                        self.rportDB[rportKey] = [c_dualIf, k_age, False, 0]
                        self.mcastDBLock.release()
                        self.log_debug("Moving to local: router port %s bridge %s dualIf %s age %s" %
                                       (k_port, k_bridge, c_dualIf, k_age))
                else:
                    # Router port is owned by this switch. Update/add entry
                    self.mcastDBLock.acquire()
                    if rportKey not in self.rportDB:
                        self.mdbFromKernWorkQueue.put(
                            (mdbsync._WORK_QUEUE_RPORT_ADD, [k_bridge, k_port]))
                    else:
                        self.rportDB[rportKey] = [c_dualIf, k_age, False, 0]
                    self.mcastDBLock.release()
                try:
                    c_rportDB.pop((k_bridge, k_port))
                except KeyError:
                    self.log_debug(" No Entry")

        for (bridge, port), [dualIf, age, remote, updated] in list(c_rportDB.items()):
            self.mdbFromKernWorkQueue.put(
                (mdbsync._WORK_QUEUE_RPORT_DELETE, [bridge, port]))

        self.SyncCurrentStateToPeer(kMdb)

    def RefreshMcasts(self):
        """
        This function execute periodically 
        Default is 4 s . The periodicity of refresh can be configured 
        a)It takes Kernel Dump 
        b)Updates the Ages of the local multicast entry . 
        c)Verifies if the entry is remote and we have seen the mcast entries more recently . 
            if yes it converts the remote entry to local entry .
        d)Sync the local entries to the peer . 
        """
        self.log_verbose("Beginning execution of the thread RefreshMcasts")
        lastAgeRefresh = int(os.times()[4])

        while self.keep_going:
            self.log_debug("RefreshMcasts ..things to do ")
            """
            Do all these things periodically
            """
            shortestRefreshAge = 255
            for br in self.Intf.GetAllBridges():
                maxAge = min(self.Intf.GetBridgeMemInterval(
                    br), self.Intf.GetBridgeQuerierInterval(br))
                if maxAge is not None and maxAge < shortestRefreshAge:
                    shortestRefreshAge = maxAge
            now = int(os.times()[4])
            if (now - lastAgeRefresh) > (shortestRefreshAge * 0.5):
                self.log_debug("Adding entry to the mdbWorkQueue %s " %
                               ("DoRefreshMcasts",))
                self.mdbFromKernWorkQueue.put(
                    (mdbsync._WORK_QUEUE_REFRESH, ()))
                lastAgeRefresh = now

            self.log_verbose("Waiting %d seconds before continuing execution of the thread RefreshMcasts"
                             % (self.Parser.getPeriodicRun(),))

            self.stopEvent.wait(self.Parser.getPeriodicRun())
            self.log_debug("Executing the thread RefreshMcasts ..thread")

        self.log_verbose("Finished execution of the thread RefreshMcasts")

    def MdbKernelMsg(self):
        """
        This is an independent thread .It is responsible for 
        processing entries on MdbWorkQueue, which may result in 
        modification to the mdb and messages sent to the peer.
        """
        self.log_verbose("Beginning execution of the thread MdbKernelMsg")
        self.clagnl.DumpMcasts()
        nlm = nlmanager.nlmanager.NetlinkManager(util.CLAG_NLM_MDB_GET_ID)

        while self.keep_going:
            self.log_verbose("Wating for an event before continuing execution "
                             "of the thread MdbKernelMsg")
            (opcode, k_params) = self.mdbFromKernWorkQueue.get()

            if self.keep_going:
                queueEmpty = False
                vlanMap = self.Intf.GetSubIfsVlanMap()
                dualIfs = self.Intf.GetDualIfs()
                peerIfs = set(self.Intf.GetSubIfsOfMaster(
                    self.peerIf)) | set([self.peerIf])
                peerlinkOfBridge = self.Intf.GetPeerlinksOfBridgeMap()
                protobuf = self.Parser.peerProtoBuf
                if protobuf:
                    msgData = peerXchange.MdbSync()
                else:
                    msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
                Entries = 0
                while not queueEmpty and self.keep_going:
                    if opcode == mdbsync._WORK_QUEUE_NOP:
                        (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()
                        continue
                    elif opcode == mdbsync._WORK_QUEUE_REFRESH:
                        self.DoRefreshMcasts(nlm)

                    elif opcode == mdbsync._WORK_QUEUE_MCAST_ADD or \
                            opcode == mdbsync._WORK_QUEUE_MCAST_DELETE:
                        self.log_debug("Processing mcast entry: %d %s" %
                                       (opcode, str(k_params)))
                        (k_bridge, k_port, k_state,
                         k_flags, k_vid, k_group) = k_params
                        # TODO Test for subinterface case What is the msg received
                        isPeerLink = True if k_port in peerIfs else False
                        igmp_fleave = True if k_flags & mdbsync._MDB_FLAGS_FAST_LEAVE else False

                        if not k_vid:
                            self.log_debug("k_vid is None Traditional Bridge")
                            (k_dualIf, k_vid) = vlanMap.get(k_port, (k_port, 0))
                        else:
                            k_dualIf = k_port

                        self.log_debug("k_vid = %s k_dualIf = %s k_port = %s " % (
                            k_vid, k_dualIf, k_port))

                        mcastKey = (k_group, k_bridge, k_port, k_vid)

                        if opcode == mdbsync._WORK_QUEUE_MCAST_ADD:
                            self.mcastDBLock.acquire()
                            if mcastKey not in self.mcastDB:
                                self.mcastDB[mcastKey] = [
                                    k_dualIf, 0, False, 0]
                                self.log_debug("New local multicast group %s bridge %s port %s vlanId %s dualIf %s" %
                                               (k_group, k_bridge, k_port, k_vid, k_dualIf))
                                # Skip syncing multicast group entries to peer when received over peerlink
                                # Skip syncing of Permanent Group Entries
                                if not isPeerLink and k_state != 1:
                                    peerlinkIntf = peerlinkOfBridge.get(k_bridge, None)
                                    self.AddPortToMulitcastGroup(k_bridge, k_vid, k_group, peerlinkIntf, k_port)
                                    if protobuf:
                                        self.AddProtoBufEntry(msgData, "group", "mcast-add",
                                                              k_dualIf, k_bridge, 0, k_vid, k_group)
                                    else:
                                        self.AddMcastXmlEntry(msgData, "mcast-add", k_dualIf,
                                                              k_bridge, 0, k_vid, k_group)
                                    Entries += 1
                                else:
                                    self.log_debug(
                                        "Mcast Entry on peerlink .Skipping remote update.")
                            # else TODO
                            self.mcastDBLock.release()
                        else:
                            # If IGMP Leave is received on:
                            # clagbond: Delete the user info
                            # peerlink: Skip delete of the entry if user exists

                            # If igmp leave is received on a peerlink and if users exists for the 
                            # mcast group in which peerlink is part of . Do not delete the entry.
                            self.mcastPeerlinkUserDBLock.acquire()
                            if isPeerLink and igmp_fleave:
                                if self.mcastPeerlinkUserDB.get(mcastKey, None):
                                    if not self.Intf.isVlanFilteringBridge(k_bridge):
                                        self.ToKernWorkQueuePut((mdbsync._MCAST_ADD, k_group, k_bridge, k_port, 0))
                                    else:
                                        self.ToKernWorkQueuePut((mdbsync._MCAST_ADD, k_group, k_bridge, k_port, k_vid))
                                    (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()     
                                    self.log_debug("Multicast Users exists skipping deleting the entry")
                                    self.mcastPeerlinkUserDBLock.release()
                                    continue
                            self.mcastPeerlinkUserDBLock.release()

                            self.log_debug("Deleting local multicast group %s bridge %s port %s vlanId %s kdualIf %s" %
                                           (k_group, k_bridge, k_port, k_vid, k_dualIf))

                            # If IGMP Leave is received on the clag bonds then delete the user from the peerlink user db.
                            self.mcastDBLock.acquire()
                            if mcastKey in self.mcastDB:
                                del self.mcastDB[mcastKey]
                            if not isPeerLink and igmp_fleave:
                                peerlinkIntf = peerlinkOfBridge.get(k_bridge, None)
                                self.DelPeerlinkMulticastUser(k_bridge, k_vid, k_group, peerlinkIntf, k_port)
                            self.mcastDBLock.release()

                            self.mcastPeerlinkUserDBLock.acquire()
                            if isPeerLink:
                                if not igmp_fleave and (mcastKey in self.mcastPeerlinkUserDB):
                                    self.log_debug("Deleting the peerlink multicast group users for group %s bridge %s vlanId %s" %
                                                        (k_group, k_bridge, k_vid))
                                    del self.mcastPeerlinkUserDB[mcastKey]
                            self.mcastPeerlinkUserDBLock.release()

                            if igmp_fleave and not isPeerLink:
                                self.log_debug("Rxed: Igmp Fast Leave Notifying Peer to Delete \
                                                    %s bridge %s port %s vlanId %s kdualIf %s" % (k_group, k_bridge, k_port, k_vid, k_dualIf))
                                if protobuf:
                                    self.AddProtoBufEntry(msgData, "group", "mcast-delete", k_dualIf,
                                                          k_bridge, 0, k_vid, k_group)
                                else:
                                    self.AddMcastXmlEntry(msgData, "mcast-delete", k_dualIf,
                                                          k_bridge, 0, k_vid, k_group)
                                Entries += 1

                        # TODO update k_age for delete case

                    elif opcode == mdbsync._WORK_QUEUE_RPORT_ADD or \
                            opcode == mdbsync._WORK_QUEUE_RPORT_DELETE:
                        self.log_debug("Processing router entry: %d %s" % (
                            opcode, str(k_params)))
                        (k_bridge, k_port) = k_params
                        k_dualIf = vlanMap.get(k_port, (k_port, None))[
                            0] if self.Intf.isSubIfName(k_port) else k_port
                        self.log_debug("Interface rtrType : %d" % (
                            self.Intf.getBridgeMcastRtrType(k_port)))

                        rportKey = (k_bridge, k_port)

                        if opcode == mdbsync._WORK_QUEUE_RPORT_ADD:
                            rportState = self.Intf.getBridgeMcastRtrType(
                                k_port)
                            self.mcastDBLock.acquire()
                            self.log_debug("New local router entry bridge %s port %s  dualIf %s" %
                                           (k_bridge, k_port, k_dualIf))
                            if rportKey not in self.rportDB:
                                self.rportDB[rportKey] = [
                                    k_dualIf, 0, False, 0]
                                # Do not sync permanent rport entries
                                if rportState != nlmanager.nlpacket.MDB.MDB_RTR_TYPE_PERM:
                                    if protobuf:
                                        self.AddProtoBufEntry(msgData, "router", "rport-add", k_dualIf, k_bridge, 0)
                                    else:
                                        self.AddRouterXmlEntry(msgData, "rport-add", k_dualIf, k_bridge, 0)
                                    Entries += 1
                            # else:TODO
                            self.mcastDBLock.release()
                        else:
                            self.log_debug("Delete router entry bridge %s port %s  dualIf %s" %
                                           (k_bridge, k_port, k_dualIf))
                            self.mcastDBLock.acquire()
                            if rportKey in self.rportDB:
                                del self.rportDB[rportKey]
                            # else: TODO
                            self.mcastDBLock.release()
                            if protobuf:
                                self.AddProtoBufEntry(msgData, "router", "rport-delete", k_dualIf, k_bridge, 0)
                            else:
                                self.AddRouterXmlEntry(msgData, "rport-delete", k_dualIf, k_bridge, 0)
                            Entries += 1
                        # TODO update k_age for delete case
                    if Entries >= mdbsync._MAX_SYNC and self.keep_going:
                        msgData = self.QueueMessage(protobuf, msgData)
                        Entries = 0

                    (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()

                if Entries and self.keep_going:
                    self.QueueMessage(protobuf, msgData)

        self.log_verbose("Finished execution of the thread MdbKernelMsg")

    def UpdateMcastsInHw(self):
        '''
        This function is an independent thread of the script. 
        '''
        mdb_cmd = {
            mdbsync._MCAST_ADD: (nlmanager.nlpacket.RTM_NEWMDB, nlmanager.nlpacket.NLM_F_CREATE |
                                 nlmanager.nlpacket.NLM_F_REPLACE, "added (mdba-mdb) to"),
            mdbsync._MCAST_REFRESH: (nlmanager.nlpacket.RTM_NEWMDB, nlmanager.nlpacket.NLM_F_CREATE |
                                     nlmanager.nlpacket.NLM_F_REPLACE, "refreshed (mdba-mdb )on"),
            mdbsync._MCAST_DELETE: (nlmanager.nlpacket.RTM_DELMDB, 0, "removed (mdba-mdb) from"),
        }

        self.log_verbose("Beginning execution of the thread UpdateMcastsInHw")
        nlm = mdb_nlmanager(self.Log)

        while self.keep_going:
            self.log_verbose(
                "Waiting for an event before continuing execution of the thread UpdateToKernel")
            (op, group, bridge, port, vlanId) = self.mdbToKernWorkQueue.get()

            if self.keep_going:
                queueEmpty = False
                mcast_params = []
                rport_add_params = []
                rport_delete_params = []

                while not queueEmpty and self.keep_going:
                    if op in [mdbsync._MCAST_ADD, mdbsync._MCAST_REFRESH, mdbsync._MCAST_DELETE]:
                        (rtm_command, nlm_flags, mdb_str) = mdb_cmd.get(
                            op, (None, None, None))
                        if rtm_command is not None:
                            bridge = self.Intf.GetIfIndex(bridge)
                            if bridge is None or port is None or group is None:
                                try:
                                    (op, group, bridge, port, vlanId) = self.mdbToKernWorkQueue.get(
                                        True, 0.001)
                                except queue.Empty:
                                    queueEmpty = True
                                continue
                            port = self.Intf.GetIfIndex(port)
                            mcast_params += [(rtm_command, nlm_flags,
                                              group, bridge, port, vlanId)]
                            self.log_debug("Mcast %s %s has been %s %s in vlan %d"
                                           % (group, bridge, mdb_str, port, vlanId))
                    elif op in [mdbsync._RPORT_ADD, mdbsync._RPORT_REFRESH]:
                        port = self.Intf.GetIfIndex(port)
                        if port is None:
                            try:
                                (op, group, bridge, port, vlanId) = self.mdbToKernWorkQueue.get(
                                    True, 0.001)
                            except queue.Empty:
                                queueEmpty = True
                            continue
                        rport_add_params += [port]
                    elif op == mdbsync._RPORT_DELETE:
                        port = self.Intf.GetIfIndex(port)
                        if port is None:
                            try:
                                (op, group, bridge, port, vlanId) = self.mdbToKernWorkQueue.get(
                                    True, 0.001)
                            except queue.Empty:
                                queueEmpty = True
                            continue
                        rport_delete_params += [port]
                    else:
                        self.Log.log_error(
                            "Unknown op code in work queue: %s" % (op,))

                    try:
                        (op, group, bridge, port,
                         vlanId) = self.mdbToKernWorkQueue.get(True, 0.001)

                    except queue.Empty:
                        queueEmpty = True

                nlm.mcast_batch(mcast_params)
                self.Intf.SetBridgeMulticastRouter(
                    nlm, rport_add_params, nlmanager.nlpacket.MDB.MDB_RTR_TYPE_TEMP)
                self.Intf.SetBridgeMulticastRouter(
                    nlm, rport_delete_params, nlmanager.nlpacket.MDB.MDB_RTR_TYPE_DISABLED)

        self.log_verbose("Finished execution of the thread UpdateMcastInHw")


class mdb_nlmanager(nlmanager.nlmanager.NetlinkManager):

    def __init__(self, Log):
        super(mdb_nlmanager, self).__init__(util.CLAG_NLM_MDB_SET_ID)
        self.Log = Log

    def tx_or_concat_message(self, total_message, mcast):
        """
        Concat several of mcasts together before TXing.
        """
        PACKET_CONCAT_SIZE = 16384
        if not total_message:
            total_message = mcast.message
        else:
            total_message += mcast.message

        if len(total_message) >= PACKET_CONCAT_SIZE:
            self.tx_nlpacket_raw(total_message)
            total_message = None

        return total_message

    def mcast_batch(self, mcasts):
        total_message = None
        for(rtm_command, nlm_flags, group, bridge, port, vlanId) in mcasts:
            debugFlags = self.Log._LOG_DEBUG_NETLINK | self.Log._LOG_DEBUG_MDB_SYNC
            debug = (self.Log.getDebug() & debugFlags) == debugFlags
            mdb = nlmanager.nlpacket.MDB(rtm_command, debug, self.Log.logger,
                                         self.Log.getHandlerStr() != "syslog", tx=True)

            mdb.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlm_flags
            mdb.family = socket.AF_BRIDGE
            mdb.body = struct.pack('Bxxxi', socket.AF_BRIDGE, bridge)
# TODO Move to netaddr
            try:
                socket.inet_aton(group)
                proto = socket.htons(nlmanager.nlpacket.ETH_P_IP)
                group = IPv4Address(group)
            except socket.error:
                proto = socket.htons(nlmanager.nlpacket.ETH_P_IPV6)
                group = IPv6Address(group)

            mdb.add_attribute(nlmanager.nlpacket.MDB.MDBA_SET_ENTRY,
                              (port, 0, 0, vlanId, group, proto))
            mdb.build_message(next(self.sequence), self.pid)
            total_message = self.tx_or_concat_message(total_message, mdb)

        if total_message:
            self.tx_nlpacket_raw(total_message)
