#! /usr/bin/python3
#-------------------------------------------------------------------------------
#
# Copyright 2014, 2020 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#


try:
    import clag.datasync
    import clag.util as util
    import copy
    import json
    import os
    import queue
    import subprocess
    import tempfile
    import threading
    import socket
    import struct
    from ipaddress import IPv4Address, IPv6Address
    import select
    import xml.etree.ElementTree as ET
    from . import clagdZebraIntf_pb2 as CZ
    import clag.clagdStream_lib as stream_lib
    import clag.clagdPeerXchange_pb2 as peerXchange
except ImportError as e:
    raise ImportError (str(e) + "- required module not found")


class mroutesync(clag.datasync.datasync):
    """
    This class handles everything required to propogate mroute entries
    between two clag peers.
    Unlike other datasyncs clag is not subscribed to rtnl_mroute notifications.
    The Mroute Updates are received via zebra.
    """
    _CLAG_ZEBRA_CONNECTED = 0
    _CLAG_ZEBRA_DISCONNECTED = 1

    # Work queue operation codes.
    _MROUTE_ADD = 0
    _MROUTE_DEL = 1
    _MROUTE_BULK_ADD = 2
    _MROUTE_BULK_DEL = 3
    _ZEBRA_STATUS_UPDATE = 4
    _CLAG_STATUS_UPDATE = 5
    _VXLAN_STATUS_UPDATE = 6

    # Mroute Data
    MROUTE_DB_COST_TO_RP = 0
    MROUTE_DB_AM_I_DR = 1
    MROUTE_DB_AM_I_DUAL_ACTIVE = 2

    _IP_ANY = "0.0.0.0"

    _ClagZebraSocketFile = "/var/run/clag-zebra.socket"

    # Max entries synced to peer in a single protobuf/XML message
    _MAX_SYNC = 500

    def __init__(self, Daemon, Intf, Parser, Log, sendQueue):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(mroutesync, self).__init__()
        #
        #  Save the parameters passed in
        #
        self.Daemon = Daemon
        self.Intf   = Intf
        self.Parser = Parser
        self.Log    = Log
        self.sendQueue = sendQueue
        self.peerIf = self.Parser.args.peerIf

        #
        #  MROUTE Database - There are two MROUTE Database, one for our switch
        #  and one for peerswitch. The database is a dictionary whose keys,values are
        #  (vrf, source, group, vni) [cost_to_rp, am_i_dr,  am_i_dual_active]
        #  The Database has been maintained only for syncing purpose between local pim
        #  and remote pim.
        #  cost_to_rp is the cost between the router and RP
        #  am_i_dr is if clag peer is a DF for corresponding (S,G) Entry
        #  am_i_dual_Active is if active-active feature is enabled on clag
        #
        self.ourMrouteDBLock    = threading.Lock()
        self.ourMrouteDB        = {}
        self.peerMrouteDBLock   = threading.Lock()
        self.peerMrouteDB       = {}

        self.ClagZebraSrvSock = None
        self.clagZebraConnDB = {}
        #Zebra Connection Socket
        self.clients = {}
        self.peer_frr_state = CZ.ZebraMlagZebraStatusUpdate.FrrState.Value('FRR_STATE_DOWN')

        # Mroute Updates from zebra are processed by this queue
        self.mrouteDataRecvQ = queue.Queue()
        util.queueAddNameMap(self.mrouteDataRecvQ, 'mrouteDataRecvQ')

        # Clag Status Notifications to Zebra are processed by this queue
        self.mrouteDataSendQ = queue.Queue()
        util.queueAddNameMap(self.mrouteDataSendQ, 'mrouteDataSendQ')

        self.keep_going = True
        self.stopEvent  = clag.clagthread.Event()

        self.notifyThread = threading.Thread(None, self.ClagZebraListenT, "ClagZebraListenT")
        self.notifyThread.daemon = True
        self.notifyThread.start()

        self.updateThread = threading.Thread(None, self.UpdateMrouteDataT, "UpdateMrouteDataT")
        self.updateThread.daemon = True
        self.updateThread.start()

        self.sendDataThread = threading.Thread(None, self.UpdateToZebraT, "UpdateToZebraT")
        self.sendDataThread.daemon = True
        self.sendDataThread.start()

    def __del__(self):
        """
        Destructor:
        """
        #
        #   Signal the update thread to wake up and exit
        #
        self.SignalThreads()
        self.JoinThreads()

    def UpdateMrouteDataT(self):
        try:
            self.UpdateMrouteData()
        except:
            self.Daemon.DumpTraceback()

    def ClagZebraListenT(self):
        try:
            self.ClagZebraListen()
        except:
            self.Daemon.DumpTraceback()

    def UpdateToZebraT(self):
        try:
            self.UpdateToZebra()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.stopEvent.set()
        try:
            self.mrouteDataRecvQ.put_nowait(None)
            util.queueSetMaxLen('mrouteDataRecvQ')
        except queue.Full:
            pass
        try:
            self.mrouteDataSendQ.put_nowait((None, None))
            util.queueSetMaxLen('mrouteDataSendQ')
        except queue.Full:
            pass

    def JoinThreads(self):
        self.sock_shut(self.ClagZebraSrvSock)
        self.updateThread.join()
        self.sendDataThread.join()
        self.notifyThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_MROUTE_SYNC, logStr)

    def log_verbose(self, logStr):
        self.Log.log_verbose(self.Log._LOG_DEBUG_MROUTE_SYNC, logStr)

    def __sendmsg__(self, inbuff):
        zebraclientSockL = list(self.clagZebraConnDB.keys())
        if zebraclientSockL:
            self.log_debug("Building Zebra Write Data ...")
            zebraclientSock = zebraclientSockL[0]
            zebraSockInfo = self.clients.get(zebraclientSock, None)
            if zebraSockInfo:
                self.log_debug("Write Data to Zebra ")
                zebraSockInfo.build_write_data(inbuff)

    def server_sock_shut(self):
        if not self.ClagZebraSrvSock:
            return
        try:
            self.log_debug("server sock %r shut" % self.ClagZebraSrvSock)
            self.ClagZebraSrvSock.close()
            os.remove(mroutesync._ClagZebraSocketFile)
        except:
            pass
        self.ClagZebraSrvSock = None

        #This should always be 1
        zebraSock = list(self.clients.keys())
        for sock in zebraSock:
            sock.close()
            del self.clients[sock]

    def zebra_client_sock_shut(self, sock):
        self.log_debug("client sock %r shut" % sock)
        if sock in self.clients:
            del self.clients[sock]
            sock.close()

    def zebra_stale_client_sock_shut(self):
        self.log_debug("Removing all the stale zebra socket if any")
        for sock, sockInfo in list(self.clients.copy().items()):
            del self.clients[sock]
            sock.close()

    def sock_shut(self, sock):
        if sock == self.ClagZebraSrvSock:
            self.server_sock_shut()
        else:
            self.zebra_client_sock_shut(sock)

    def MrouteSendQueuePut(self, element):
        if self.keep_going:
            self.log_debug("The element is %s" % str(element))
            self.mrouteDataSendQ.put(element)
            util.queueSetMaxLen('mrouteDataSendQ')

    def SetClagZebraConn(self, zebraclientSock, status):
        self.clagZebraConnDB = {}
        if status == mroutesync._CLAG_ZEBRA_CONNECTED:
            self.clagZebraConnDB[zebraclientSock] = "Connected"
        elif status == mroutesync._CLAG_ZEBRA_DISCONNECTED:
            self.ClearLocalInfo()
        self.UpdateLocalZebraStatus()

    def GetObjectSize(self):
        mrouteObj = {}
        self.ourMrouteDBLock.acquire()
        mrouteObj['ourMrouteDB'] = util.getSize(self.ourMrouteDB)
        self.ourMrouteDBLock.release()
        self.peerMrouteDBLock.acquire()
        mrouteObj['peerMrouteDB'] = util.getSize(self.peerMrouteDB)
        self.peerMrouteDBLock.release()
        return mrouteObj

    def UpdateLocalZebraStatus(self):
        if list(self.clagZebraConnDB.keys()):
            frrState = CZ.ZebraMlagZebraStatusUpdate.FrrState.Value('FRR_STATE_UP')
        else:
            frrState = CZ.ZebraMlagZebraStatusUpdate.FrrState.Value('FRR_STATE_DOWN')
        protobuf = self.Parser.peerProtoBuf
        if protobuf:
            msgData = peerXchange.MrouteSync()
            self.AddZebraStatusEntry(msgData, frrState)
        else:
            msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
            subroot = ET.SubElement(msgData, "mroute")
            self.AddZebraStatusXmlEntry(subroot, frrState)
        self.QueueMessage(protobuf, msgData)

    def EncodedOurMrouteStr(self):
        """
        Return local mroute Entries in human readable form
        """
        mrouteDB = []
        self.ourMrouteDBLock.acquire()
        for (vrf_name, source_ip, group_ip, vni_id, intf_name) ,params in list(self.ourMrouteDB.copy().items()):
            mroute = {}
            mroute['vrf_name'] = vrf_name
            mroute['source_ip'] = source_ip
            mroute['group_ip'] = group_ip
            mroute['vni_id'] = vni_id
            mroute['intf_name'] = intf_name
            mroute['cost_to_rp'] = params[self.MROUTE_DB_COST_TO_RP]
            mroute['am_i_DR'] = params[self.MROUTE_DB_AM_I_DR]
            mroute['am_i_Dual_active'] = params[self.MROUTE_DB_AM_I_DUAL_ACTIVE]
            mrouteDB.append(mroute)
        self.ourMrouteDBLock.release()
        return json.dumps(mrouteDB, sort_keys=True, indent=4)

    def EncodedPeerMrouteStr(self):
        """
        Return peer information in human-readable form
        """
        mrouteDB = []
        self.peerMrouteDBLock.acquire()
        for (vrf_name, source_ip, group_ip, vni_id, intf_name)  ,params in list(self.peerMrouteDB.copy().items()):
            mroute = {}
            mroute['vrf_name'] = vrf_name
            mroute['source_ip'] = source_ip
            mroute['group_ip'] = group_ip
            mroute['vni_id'] = vni_id
            mroute['intf_name'] = intf_name
            mroute['cost_to_rp'] = params[self.MROUTE_DB_COST_TO_RP]
            mroute['am_i_DR'] = params[self.MROUTE_DB_AM_I_DR]
            mroute['am_i_Dual_active'] = params[self.MROUTE_DB_AM_I_DUAL_ACTIVE]
            mrouteDB.append(mroute)
        self.peerMrouteDBLock.release()
        return json.dumps(mrouteDB, sort_keys=True, indent=4)

    def EncodedZebraStatusStr(self):
        """
        Return Local and Remote zebra status
        """
        zebraStatus = {}
        zebraStatus['local'] = "Connected" if list(self.clagZebraConnDB.keys()) else "Disconnected"
        zebraStatus['remote'] = "Connected" if self.peer_frr_state == 2 else "Disconnected"
        return json.dumps(zebraStatus, sort_keys=True, indent=4)

    def CollectLocalInfo(self, nlm):
        """
        This function is called on a periodic interval to gather up any
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.
        """
        pass

    def QueueMessage(self, protobuf, msgData):
        if protobuf:
            return self.QueueProtoBufMessage(msgData)
        else:
            return self.QueueXmlMessage(msgData)

    def QueueXmlMessage(self, msgData):
        xmlStr = ET.tostring(msgData)
        self.sendQueue.put((None, xmlStr))
        msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
        return msgData

    def QueueProtoBufMessage(self, msgData):
        if not msgData.ByteSize():
            return peerXchange.MrouteSync()
        self.Log.log_debug(self.Log._LOG_DEBUG_INFORM_PEER, "TXing mroute data: %s" % str(msgData))
        msgHdr = peerXchange.Header()
        msgHdr.type = peerXchange.Header.MessageType.Value('CLAG_MROUTE_SYNC')
        msgHdr.data = msgData.SerializeToString()
        self.sendQueue.put((msgHdr, None))
        msgData = peerXchange.MrouteSync()
        return msgData

    def MakeProtoBufMessage(self):
        """
        Writes the local information as a protocol buffer message.
        """
        if self.Parser.syncDoneToPeer:
            return

        msgData = peerXchange.MrouteSync()
        Entries = 0
        self.ourMrouteDBLock.acquire()
        for mrouteKey, mrouteValue in self.ourMrouteDB.items():
            self.AddMrouteProtoBufEntry(msgData, "mroute-add", mrouteKey, mrouteValue)
            Entries += 1
            if Entries >= mroutesync._MAX_SYNC and self.keep_going:
                Entries = self.QueueProtoBufMessage(msgData)
        self.ourMrouteDBLock.release()
        if Entries and self.keep_going:
            self.QueueProtoBufMessage(msgData)

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML for sending to the peer on
        periodic basis.
        """
        if self.Parser.syncDoneToPeer:
            return
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
        xmlEntries = 0
        subroot = ET.SubElement(root, "mroute")
        root11 = ET.SubElement(subroot, "mroute-bulk-add")
        self.ourMrouteDBLock.acquire()
        for mrouteKey, mrouteValue in list(self.ourMrouteDB.items()):
            self.AddMrouteXmlEntry(root11, "mroute-add", mrouteKey, mrouteValue)
            xmlEntries += 1
            if xmlEntries >= mroutesync._MAX_SYNC and self.keep_going:
                root = self.QueueXmlMessage(root)
                subroot = ET.SubElement(root, "mroute")
                root11 = ET.SubElement(subroot, "mroute-bulk-add")
                xmlEntries = 0
        self.ourMrouteDBLock.release()
        if xmlEntries and self.keep_going:
            self.QueueXmlMessage(root)

    def GetWorkQueueDelayed(self, delay=0.001):
        '''
        Mroute Rx/Tx Threads waits infinitely until an entry is available in the Queue.
        Once an entry is available  the subsequent entries are dequeued
        with a timeout of 1ms until the queue is empty.
        '''
        queueEmpty = False
        try:
            (inbuff) = self.mrouteDataRecvQ.get(True, delay)
        except queue.Empty:
            inbuff = None
            queueEmpty = True
        return (inbuff, queueEmpty)

    def SyncCurrentStateToPeer(self):
        """
        Syncs the local mroute entries to the peer
        """
        pass

    def UpdateOurMrouteDb(self, mrouteKey, mrouteInfo):
        """
        Updates the Local Mroute Info
        """
        self.ourMrouteDBLock.acquire()
        if mrouteKey not in self.ourMrouteDB \
            or self.ourMrouteDB[mrouteKey] != mrouteInfo:
            self.ourMrouteDB[mrouteKey] = mrouteInfo
        self.ourMrouteDBLock.release()

    def UpdatePeerMrouteDb(self, mrouteKey, mrouteInfo):
        """
        Updates the Peer Mroute Info
        """
        self.peerMrouteDBLock.acquire()
        if mrouteKey not in self.peerMrouteDB or self.peerMrouteDB[mrouteKey] != mrouteInfo:
            self.peerMrouteDB[mrouteKey] = mrouteInfo
        self.peerMrouteDBLock.release()

    def ParseProtoBufMessage(self, msgHdr):
        """
        Reads the protobuf information sent from the peer.
        """
        self.log_verbose("Beginning ParseProtoBufMessage")
        msgData = peerXchange.MrouteSync()
        msgData.ParseFromString(msgHdr.data)
        self.peerMrouteDBLock.acquire()
        mrouteAddEntries = []
        mrouteDelEntries = []
        mrouteElement = []
        for entry in msgData.entry:
            if entry.op == "mroute-add" or entry.op == "mroute-del":
                r_vrf = entry.vrf
                r_source_ip = entry.sourceIp
                r_group_ip = entry.groupIp
                r_vni_id = entry.vni
                r_intf_name = entry.iface
                if entry.op == "mroute-add":
                    r_mrouteKey = (r_vrf, r_source_ip, r_group_ip, r_vni_id, r_intf_name)
                    r_cost_to_rp = entry.costRp
                    r_am_i_DR = entry.DR
                    r_am_i_Dual_active = entry.dualActive
                    r_mrouteInfo = [r_cost_to_rp, r_am_i_DR, r_am_i_Dual_active]
                    if r_mrouteKey not in self.peerMrouteDB \
                        or self.peerMrouteDB[r_mrouteKey] != r_mrouteInfo:
                        self.log_debug("Adding Peer Mroute DB")
                        self.peerMrouteDB[r_mrouteKey] = r_mrouteInfo
                    mrouteElement = list(r_mrouteKey) + r_mrouteInfo
                    mrouteAddEntries.append(mrouteElement)
                else:
                    r_mrouteKey = (r_vrf, r_source_ip, r_group_ip, r_vni_id, r_intf_name)
                    if r_mrouteKey in self.peerMrouteDB:
                        del self.peerMrouteDB[r_mrouteKey]
                    mrouteElement = list(r_mrouteKey)
                    mrouteDelEntries.append(mrouteElement)
            elif entry.op == "zebra-status" and entry.frrState:
                self.peer_frr_state = int(entry.frrState)
                mrouteElement = [ self.peer_frr_state ]

        if list(self.clagZebraConnDB.keys()):
            if mrouteAddEntries:
                self.MrouteSendQueuePut((mroutesync._MROUTE_BULK_ADD, mrouteAddEntries))
            elif mrouteDelEntries:
                self.MrouteSendQueuePut((mroutesync._MROUTE_BULK_DEL, mrouteDelEntries))
            elif mrouteElement:
                self.MrouteSendQueuePut((mroutesync._ZEBRA_STATUS_UPDATE, mrouteElement))
        self.peerMrouteDBLock.release()
        self.log_verbose("Ending ParseProtoBufMessage")
        return msgData

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.
        root - The root element of the XML Tree
        <mroute>
            <pim-data svi_state="0" switchd_state="0" />
        </mroute>
        <mroute>
            <mroute-bulk-add>
                <mroute-add am_i_DR="False" am_i_Dual_active="True" \
                        cost_to_rp="0" group_ip="239.1.1.100" intf_name="" source_ip="27.0.0.11" \
                            vni_id="2" vrf="default" />
                <mroute-add am_i_DR="False" am_i_Dual_active="True" \
                        cost_to_rp="0" group_ip="239.1.1.100" intf_name="" source_ip="27.0.0.12" \
                            vni_id="2" vrf="default" />
            </mroute-bulk-add>
        </mroute>
        <mroute>
            <mroute-bulk-del>
                <mroute-add am_i_DR="False" am_i_Dual_active="True" \
                        cost_to_rp="0" group_ip="239.1.1.100" intf_name="" source_ip="27.0.0.11" \
                            vni_id="2" vrf="default" />
                <mroute-del am_i_DR="False" am_i_Dual_active="True" \
                        cost_to_rp="0" group_ip="239.1.1.100" intf_name="" source_ip="27.0.0.12" \
                            vni_id="2" vrf="default" />
            </mroute-bulk-del>
        </mroute>
        """

        self.peerMrouteDBLock.acquire()
        mrouteEntries = []
        mrouteData = root.find('mroute')
        mrouteDataL = mrouteData.getchildren() if mrouteData else []

        for mroutebulkData in mrouteDataL:
            mrouteAddEntries = []
            mrouteDelEntries = []
            if mroutebulkData.tag == "mroute-bulk-add":
                for mroute in mroutebulkData.getchildren():
                    r_vrf = mroute.get('vrf')
                    r_source_ip = mroute.get('source_ip')
                    r_group_ip = mroute.get('group_ip')
                    r_vni_id = mroute.get('vni_id')
                    r_intf_name = mroute.get('intf_name')
                    if r_vni_id is not None:
                        r_vni_id = int(r_vni_id)
                    r_mrouteKey = (r_vrf, r_source_ip, r_group_ip, r_vni_id, r_intf_name)
                    r_cost_to_rp = mroute.get('cost_to_rp')
                    if r_cost_to_rp is not None:
                        r_cost_to_rp = int(r_cost_to_rp)
                    r_am_i_DR = mroute.get('am_i_DR')
                    r_am_i_DR = True if r_am_i_DR == "True" else False
                    r_am_i_Dual_active = mroute.get('am_i_Dual_active')
                    r_am_i_Dual_active = True if r_am_i_Dual_active == "True" else False
                    r_mrouteInfo = [r_cost_to_rp, r_am_i_DR, r_am_i_Dual_active]
                    if r_mrouteKey not in self.peerMrouteDB \
                            or self.peerMrouteDB[r_mrouteKey] != r_mrouteInfo:
                        self.log_debug("Adding Peer Mroute DB")
                        self.peerMrouteDB[r_mrouteKey] = r_mrouteInfo
                    mrouteElement = list(r_mrouteKey) + r_mrouteInfo
                    mrouteAddEntries.append(mrouteElement)
                if list(self.clagZebraConnDB.keys()) and mrouteAddEntries:
                    self.MrouteSendQueuePut((mroutesync._MROUTE_BULK_ADD, mrouteAddEntries))

            elif mroutebulkData.tag == "mroute-bulk-del":
                for mroute in mroutebulkData.getchildren():
                    r_vrf = mroute.get('vrf')
                    r_source_ip = mroute.get('source_ip')
                    r_group_ip = mroute.get('group_ip')
                    r_vni_id = mroute.get('vni_id')
                    r_intf_name = mroute.get('intf_name')
                    if r_vni_id is not None:
                        r_vni_id = int(r_vni_id)
                    r_mrouteKey = (r_vrf, r_source_ip, r_group_ip, r_vni_id, r_intf_name)
                    if r_mrouteKey in self.peerMrouteDB:
                        del self.peerMrouteDB[r_mrouteKey]
                    mrouteElement = list(r_mrouteKey)
                    mrouteDelEntries.append(mrouteElement)
                if list(self.clagZebraConnDB.keys()) and mrouteDelEntries:
                    self.MrouteSendQueuePut((mroutesync._MROUTE_BULK_DEL, mrouteDelEntries))

            elif mroutebulkData.tag == "zebra-status":
                r_frr_state = mroutebulkData.get('frr_state')
                if r_frr_state is not None:
                    self.peer_frr_state = int(r_frr_state)
                    mrouteElement = [self.peer_frr_state]
                    if list(self.clagZebraConnDB.keys()):
                        self.MrouteSendQueuePut((mroutesync._ZEBRA_STATUS_UPDATE, mrouteElement))
            else:
                self.log_debug("Unknown Mroute Xml Tag")
        self.peerMrouteDBLock.release()

    def DualIfChange(self, dualIf, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        pass

    def MatchIdIfChange(self, dualIf, enable):
        """
        Called when an interface gets or loses a matching clag-id on the peer
        """
        pass

    def ClearPeerInfo(self):
        """
        Removes Information which has been obtained from the peer
        """
        #Verify if0 case
        self.peerMrouteDBLock.acquire()
        peerMrouteEntries = []
        numEntries = 0
        for mrouteKey, mrouteValue in list(self.peerMrouteDB.copy().items()):
            self.log_debug("Deleting mrouteKey %s" % (str(mrouteKey)))
            self.peerMrouteDB.pop(mrouteKey, None)
            mrouteElement = list(mrouteKey) + mrouteValue
            peerMrouteEntries.append(mrouteElement)
            numEntries += 1
            #Need to decide the threshold
            if numEntries > 200 and list(self.clagZebraConnDB.keys()):
                self.MrouteSendQueuePut((mroutesync._MROUTE_BULK_DEL, peerMrouteEntries))
                peerMrouteEntries = []
                numEntries = 0
        self.peerMrouteDBLock.release()
        if list(self.clagZebraConnDB.keys()) and peerMrouteEntries:
            self.MrouteSendQueuePut((mroutesync._MROUTE_BULK_DEL, peerMrouteEntries))

    def ClearLocalInfo(self):
        """
        Removes Information which has been learnt locally
        """
        Entries = 0
        protobuf = self.Parser.peerProtoBuf
        if protobuf:
            msgData = peerXchange.MrouteSync()
        else:
            msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
            subroot = ET.SubElement(msgData, "mroute")
            root11 = ET.SubElement(subroot, "mroute-bulk-del")
        self.ourMrouteDBLock.acquire()
        for mrouteKey, mrouteValue in list(self.ourMrouteDB.copy().items()):
            if protobuf:
                self.AddMrouteProtoBufEntry(msgData, "mroute-del", mrouteKey, None)
            else:
                self.AddMrouteXmlEntry(root11, "mroute-del", mrouteKey, None)
            Entries += 1
            if mrouteKey in self.ourMrouteDB:
                del self.ourMrouteDB[mrouteKey]
            if Entries >= mroutesync._MAX_SYNC and self.keep_going:
                msgData = self.QueueMessage(protobuf, msgData)
                if not protobuf:
                    subroot = ET.SubElement(msgData, "mroute")
                    root11 = ET.SubElement(subroot, "mroute-bulk-del")
                Entries = 0
        self.ourMrouteDBLock.release()
        if Entries and self.keep_going:
            self.QueueMessage(protobuf, msgData)

    def ZebraNotifyPeerZebraStatusUpdate(self):
        """
        Notifies Peer Zebra Connection Status to Zebra
        """
        zebraInfo = [ self.peer_frr_state ]
        self.MrouteSendQueuePut((mroutesync._ZEBRA_STATUS_UPDATE, zebraInfo))

    def ZebraNotifyClagStatusUpdate(self):
        """
        Any change in clag params will be notified to zebra
        """
        peerlink = self.Parser.args.peerIf
        role = self.Parser.GetCombinedRole()
        state = self.Intf.isPeerAlive()
        ClagParams = [peerlink, role, state]
        self.MrouteSendQueuePut((mroutesync._CLAG_STATUS_UPDATE, ClagParams))

    def ZebraNotifyVxlanUpdate(self, anycast_ip, local_ip):
        """
        Vxlan Local Ip /Anycast Ip Updates are notified here
        """
        VxlanParams = [anycast_ip, local_ip]
        self.MrouteSendQueuePut((mroutesync._VXLAN_STATUS_UPDATE, VxlanParams))

    def AddMrouteProtoBufEntry(self, data, op, mrouteKey, mrouteValue):
        """
        Adds an entry to the protocol buffer MrouteSync message
        """
        entry = data.entry.add()
        (vrf_name, source_ip, group_ip, vni_id, intf_name) = mrouteKey
        if op == "mroute-add":
            [cost_to_rp, am_i_DR, am_i_Dual_active] = mrouteValue
            self.log_debug(" vni is %d" % (vni_id))
            entry.costRp = cost_to_rp
            entry.DR = am_i_DR
            entry.dualActive = am_i_Dual_active
        entry.op = op
        entry.vrf = vrf_name
        entry.sourceIp = str(source_ip)
        entry.groupIp = str(group_ip)
        entry.vni = vni_id
        entry.iface = str(intf_name)

    def AddMrouteXmlEntry(self, element, op, mrouteKey, mrouteValue):
        (vrf_name, source_ip, group_ip, vni_id, intf_name) = mrouteKey
        if op == "mroute-add":
            [cost_to_rp, am_i_DR, am_i_Dual_active] = mrouteValue
            self.log_debug(" vni is %d" % (vni_id))
            mrouteAttrib = { "vrf":vrf_name, "source_ip":str(source_ip), "group_ip":str(group_ip),
                             "vni_id":str(vni_id), "intf_name":str(intf_name), "cost_to_rp":str(cost_to_rp),
                             "am_i_DR":str(am_i_DR), "am_i_Dual_active":str(am_i_Dual_active)}
        elif op == "mroute-del":
            mrouteAttrib = { "vrf":vrf_name, "source_ip":str(source_ip), \
                                    "group_ip":str(group_ip), "vni_id":str(vni_id), "intf_name":intf_name}
        ET.SubElement(element, op, mrouteAttrib)

    def AddZebraStatusEntry(self, data, frr_state):
        entry = data.entry.add()
        entry.op = "zebra-status"
        entry.frrState = str(frr_state)

    def AddZebraStatusXmlEntry(self, element, frr_state):
        zebraStatusAttrib = {"frr_state":str(frr_state)}
        ET.SubElement(element, "zebra-status", zebraStatusAttrib)

    def DecodePbMrouteAddData(self, msgData, msg):
        vrf_name = msg.vrf_name
        source_ip = str(IPv4Address(msg.source_ip)) if msg.source_ip else mroutesync._IP_ANY
        group_ip = str(IPv4Address(msg.group_ip)) if msg.group_ip else mroutesync._IP_ANY
        vni_id = msg.vni_id
        cost_to_rp = msg.cost_to_rp
        am_i_DR = msg.am_i_DR
        am_i_Dual_active = msg.am_i_Dual_active
        intf_name = msg.intf_name
        mrouteKey = (vrf_name, source_ip, group_ip, vni_id, intf_name)
        mrouteInfo = [cost_to_rp, am_i_DR, am_i_Dual_active]
        self.log_debug("DecodePbMrouteAddData %s %s " % (str(mrouteKey), str(mrouteInfo)))
        self.UpdateOurMrouteDb(mrouteKey, mrouteInfo)
        self.AddMrouteProtoBufEntry(msgData, "mroute-add", mrouteKey, mrouteInfo)
        return 1

    def DecodePbMrouteAddXmlData(self, root, msg):
        vrf_name = msg.vrf_name
        source_ip = str(IPv4Address(msg.source_ip)) if msg.source_ip else mroutesync._IP_ANY
        group_ip = str(IPv4Address(msg.group_ip)) if msg.group_ip else mroutesync._IP_ANY
        vni_id = msg.vni_id
        cost_to_rp = msg.cost_to_rp
        am_i_DR = msg.am_i_DR
        am_i_Dual_active = msg.am_i_Dual_active
        intf_name = msg.intf_name
        mrouteKey = (vrf_name, source_ip, group_ip, vni_id, intf_name)
        mrouteInfo = [cost_to_rp, am_i_DR, am_i_Dual_active]
        self.log_debug("DecodePbMrouteAddXmlData %s %s " % (str(mrouteKey), str(mrouteInfo)))
        self.UpdateOurMrouteDb(mrouteKey, mrouteInfo)
        self.AddMrouteXmlEntry(root, "mroute-add", mrouteKey, mrouteInfo)
        return 1

    def EncodePbMrouteAddData(self, msgHdr, mrouteInfo):
        (vrf_name, source_ip, group_ip, vni_id, intf_name,\
            cost_to_rp, am_i_DR, am_i_Dual_active) = mrouteInfo
        msgHdr.type = CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_MROUTE_ADD')
        msgData = CZ.ZebraMlagMrouteAdd()
        msgData.vrf_name = vrf_name
        source_ip = int(IPv4Address(source_ip))
        msgData.source_ip = source_ip
        group_ip = int(IPv4Address(group_ip))
        msgData.group_ip = group_ip
        msgData.vni_id = vni_id
        msgData.intf_name = intf_name
        msgData.cost_to_rp = cost_to_rp
        msgData.am_i_DR = am_i_DR
        msgData.am_i_Dual_active = am_i_Dual_active
        serializedPld = msgData.SerializeToString()
        msgHdr.data = serializedPld

    def EncodePbMrouteBulkAddData(self, msgHdr, mrouteElements):
        mrouteEntryObjs = []
        msgData = CZ.ZebraMlagMrouteAddBulk()
        for i in range(0, len(mrouteElements)):
            mrouteEntry = mrouteElements[i]
            mrouteEntryObjs.append(CZ.ZebraMlagMrouteAdd(vrf_name = mrouteEntry[0], \
                        source_ip = int(IPv4Address(mrouteEntry[1])),
                        group_ip = int(IPv4Address(mrouteEntry[2])), \
                        vni_id = mrouteEntry[3], intf_name = mrouteEntry[4], cost_to_rp = mrouteEntry[5], \
                        am_i_DR = mrouteEntry[6], am_i_Dual_active = mrouteEntry[7]))
        msgData  = CZ.ZebraMlagMrouteAddBulk(mroute_add = mrouteEntryObjs)
        serializedPld = msgData.SerializeToString()
        msgHdr.data = serializedPld

    def EncodePbMrouteBulkDelData(self, msgHdr, mrouteElements):
        mrouteEntryObjs = []
        msgData = CZ.ZebraMlagMrouteDelBulk()
        for i in range(0, len(mrouteElements)):
            mrouteEntry = mrouteElements[i]
            mrouteEntryObjs.append(CZ.ZebraMlagMrouteDel(vrf_name = mrouteEntry[0], \
                        source_ip = int(IPv4Address(mrouteEntry[1])),
                        group_ip = int(IPv4Address(mrouteEntry[2])), \
                        vni_id = mrouteEntry[3], intf_name = mrouteEntry[4]))
        msgData  = CZ.ZebraMlagMrouteDelBulk(mroute_del = mrouteEntryObjs)
        serializedPld = msgData.SerializeToString()
        msgHdr.data = serializedPld

    def DecodePbMrouteBulkAddData(self, msgData, msg):
        Entries = 0
        for i in range(0, len(msg.mroute_add)):
            Entries += self.DecodePbMrouteAddData(msgData, msg.mroute_add[i])
        return Entries

    def DecodePbMrouteBulkAddXmlData(self, root, msg):
        xmlEntries = 0
        root11 = ET.SubElement(root, "mroute-bulk-add")
        for i in range(0, len(msg.mroute_add)):
            xmlEntry = self.DecodePbMrouteAddXmlData(root11, msg.mroute_add[i])
            xmlEntries += xmlEntry
        return xmlEntries

    def DecodePbMrouteDelData(self, msgData, msg):
        vrf_name = msg.vrf_name
        source_ip = str(IPv4Address(msg.source_ip)) if msg.source_ip else mroutesync._IP_ANY
        group_ip = str(IPv4Address(msg.group_ip)) if msg.group_ip else mroutesync._IP_ANY
        vni_id = msg.vni_id
        intf_name = msg.intf_name
        mrouteKey = (vrf_name, source_ip, group_ip, vni_id, intf_name)
        self.ourMrouteDBLock.acquire()
        self.log_debug("Deleting the data for mrouteKey = %s" % (str(mrouteKey)))
        self.ourMrouteDB.pop(mrouteKey, None)
        self.ourMrouteDBLock.release()
        self.AddMrouteProtoBufEntry(msgData, "mroute-del", mrouteKey, None)
        return 1

    def DecodePbMrouteDelXmlData(self, root, msg):
        vrf_name = msg.vrf_name
        source_ip = str(IPv4Address(msg.source_ip)) if msg.source_ip else mroutesync._IP_ANY
        group_ip = str(IPv4Address(msg.group_ip)) if msg.group_ip else mroutesync._IP_ANY
        vni_id = msg.vni_id
        intf_name = msg.intf_name
        mrouteKey = (vrf_name, source_ip, group_ip, vni_id, intf_name)
        self.ourMrouteDBLock.acquire()
        self.log_debug("Deleting the data for mrouteKey = %s" % (str(mrouteKey)))
        self.ourMrouteDB.pop(mrouteKey, None)
        self.ourMrouteDBLock.release()
        self.AddMrouteXmlEntry(root, "mroute-del", mrouteKey, None)
        return 1

    def DecodePbMrouteBulkDelData(self, msgData, msg):
        Entries = 0
        for i in range(0, len(msg.mroute_del)):
            Entries += self.DecodePbMrouteDelData(msgData, msg.mroute_del[i])
        return Entries

    def DecodePbMrouteBulkDelXmlData(self, root, msg):
        xmlEntries = 0
        root11 = ET.SubElement(root, "mroute-bulk-del")
        for i in range(0, len(msg.mroute_del)):
            xmlEntry = self.DecodePbMrouteDelXmlData(root11, msg.mroute_del[i])
            xmlEntries += xmlEntry
        return xmlEntries

    def EncodePbZebraStatusUpdate(self, msgHdr, zebraStatusInfo):
        frr_state = zebraStatusInfo[0]
        msgData = CZ.ZebraMlagZebraStatusUpdate()
        msgData.peer_frrstate = frr_state
        serializedPld = msgData.SerializeToString()
        msgHdr.data = serializedPld

    def EncodePbClagParams(self, msgHdr, clagInfo):
        (peerlink, role, state) = clagInfo
        my_role = CZ.ZebraMlagStatusUpdate.ClagRole.Value('CLAG_ROLE_PRIMARY') \
            if role == "primary" else CZ.ZebraMlagStatusUpdate.ClagRole.Value('CLAG_ROLE_SECONDARY')
        clag_state = CZ.ZebraMlagStatusUpdate.ClagState.Value('CLAG_STATE_UP') \
            if state == True else CZ.ZebraMlagStatusUpdate.ClagState.Value('CLAG_STATE_DOWN')
        msgData = CZ.ZebraMlagStatusUpdate()
        msgData.peerlink = peerlink
        msgData.my_role = my_role
        msgData.clag_state = clag_state
        serializedPld = msgData.SerializeToString()
        msgHdr.data = serializedPld

    def EncodePbVxlanParams(self, msgHdr, vxlanInfo):
        (anycast_ip, local_ip) = vxlanInfo
        msgData = CZ.ZebraMlagVxlanUpdate()
        if anycast_ip:
            msgData.anycast_ip = int(IPv4Address(anycast_ip))
        if local_ip:
            msgData.local_ip = int(IPv4Address(local_ip))
        serializedPld = msgData.SerializeToString()
        msgHdr.data = serializedPld

    def ReplayPeerDb(self):
        self.peerMrouteDBLock.acquire()
        peerMrouteEntries = []
        numEntries = 0
        for mrouteKey, mrouteValue in list(self.peerMrouteDB.copy().items()):
            mrouteElement = list(mrouteKey) + mrouteValue
            peerMrouteEntries.append(mrouteElement)
            numEntries += 1
            #Need to decide the threshold
            if numEntries > 200 and list(self.clagZebraConnDB.keys()):
                self.MrouteSendQueuePut((mroutesync._MROUTE_BULK_ADD, peerMrouteEntries))
                peerMrouteEntries = []
                numEntries = 0
        self.peerMrouteDBLock.release()
        if list(self.clagZebraConnDB.keys()) and peerMrouteEntries:
            self.MrouteSendQueuePut((mroutesync._MROUTE_BULK_ADD, peerMrouteEntries))

    def HandleZebraRegister(self, msg):
        self.log_debug(" Zebra Register")
        self.ZebraNotifyPeerZebraStatusUpdate()
        self.ZebraNotifyClagStatusUpdate()
        #Update the localIp And Anycast Ip to Zebra
        self.ZebraNotifyVxlanUpdate(self.Parser.args.vxlanAnycast, \
                                    self.Parser.localVtepIp)
        #Replay the PeerDB on Zebra Registration
        self.ReplayPeerDb()

    def UpdateMrouteData(self):
        """
        This is an independent thread. It is responsible for processing the mroute data
        received from the client and Updating to the peer.
        """
        self.log_verbose("Beginning execution of the thread UpdateMrouteData")
        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution " \
                    "of the thread UpdateMrouteData")
            inbuff = self.mrouteDataRecvQ.get()
            if self.keep_going:
                queueEmpty = False
                protobuf = self.Parser.peerProtoBuf
                if protobuf:
                    msgData = peerXchange.MrouteSync()
                else:
                    msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
                    subroot = ET.SubElement(msgData, "mroute")
                Entries = 0
                while not queueEmpty and self.keep_going:
                    try:
                        zebraMlagMsgHdr = CZ.ZebraMlag_Header()
                        zebraMlagMsgHdr.ParseFromString(inbuff)
                        zebraMlagMsgPld = zebraMlagMsgHdr.data
                    except CZ._message.DecodeError:
                        (inbuff, queueEmpty) = self.GetWorkQueueDelayed()
                        continue

                    if zebraMlagMsgHdr.type \
                        == CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_REGISTER'):
                        self.log_debug("Rxed Zebra Register")
                        self.HandleZebraRegister(zebraMlagMsgPld)
                        (inbuff, queueEmpty) = self.GetWorkQueueDelayed()
                        continue


                    elif zebraMlagMsgHdr.type \
                        == CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_MROUTE_ADD_BULK'):
                        try:
                            self.log_debug("Rxed Mroute Add Bulk Data")
                            zebraMlagMsg = CZ.ZebraMlagMrouteAddBulk()
                            zebraMlagMsg.ParseFromString(zebraMlagMsgPld)
                        except CZ._message.DecodeError:
                            (inbuff, queueEmpty) = self.GetWorkQueueDelayed()
                            continue
                        if protobuf:
                            Entries += self.DecodePbMrouteBulkAddData(msgData, zebraMlagMsg)
                        else:
                            Entries += self.DecodePbMrouteBulkAddXmlData(subroot, zebraMlagMsg)
                        self.log_debug("Number of Add Entries = %d" % (Entries))
                    elif zebraMlagMsgHdr.type \
                        == CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_MROUTE_DEL_BULK'):
                        try:
                            self.log_debug("Rxed Mroute Del Bulk Data")
                            zebraMlagMsg = CZ.ZebraMlagMrouteDelBulk()
                            zebraMlagMsg.ParseFromString(zebraMlagMsgPld)
                        except CZ._message.DecodeError:
                            (inbuff, queueEmpty) = self.GetWorkQueueDelayed()
                            continue
                        if protobuf:
                            Entries += self.DecodePbMrouteBulkDelData(msgData, zebraMlagMsg)
                        else:
                            Entries += self.DecodePbMrouteBulkDelXmlData(subroot, zebraMlagMsg)
                        self.log_debug("Number of Del Entries = %d" % (Entries))
                    else:
                        self.log_debug("Unknown Message Type Received from frr")
                        (inbuff, queueEmpty) = self.GetWorkQueueDelayed()
                        continue

                    if Entries >= mroutesync._MAX_SYNC and self.keep_going:
                        self.log_debug("Total Entries exceeds 2000 = %d" % (Entries))
                        msgData = self.QueueMessage(protobuf, msgData)
                        if not protobuf:
                            subroot = ET.SubElement(msgData, "mroute")
                        Entries = 0
                    (inbuff, queueEmpty) = self.GetWorkQueueDelayed()

                self.log_debug("Total Entries = %d" %(Entries))
                if Entries and self.keep_going:
                    self.QueueMessage(protobuf, msgData)
        self.log_verbose("Finished execution of the thread MrouteDataRecv")

    def UpdateToZebra(self):
        """
        This is an independent Thread , It is responsible for processing
        encoding the clag Updates and Updates from remote pim to protobuff and
        sending it over client socket.
        """
        self.log_verbose("Beginning execution of the thread UpdateToZebra")
        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution of the thread")
            (op, params) = self.mrouteDataSendQ.get()
            if self.keep_going:
                queueEmpty = False
                while not queueEmpty and self.keep_going:
                    msgHdr = CZ.ZebraMlag_Header()
                    if op == mroutesync._MROUTE_BULK_ADD:
                        msgHdr.type = CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_MROUTE_ADD_BULK')
                        self.EncodePbMrouteBulkAddData(msgHdr, params)
                    elif op == mroutesync._MROUTE_BULK_DEL:
                        msgHdr.type = CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_MROUTE_DEL_BULK')
                        self.EncodePbMrouteBulkDelData(msgHdr, params)
                    elif op == mroutesync._ZEBRA_STATUS_UPDATE:
                        msgHdr.type = CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_ZEBRA_STATUS_UPDATE')
                        self.EncodePbZebraStatusUpdate(msgHdr, params)
                    elif op == mroutesync._CLAG_STATUS_UPDATE:
                        msgHdr.type = CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_STATUS_UPDATE')
                        self.EncodePbClagParams(msgHdr, params)
                    elif op == mroutesync._VXLAN_STATUS_UPDATE:
                        msgHdr.type = CZ.ZebraMlag_Header.MessageType.Value('ZEBRA_MLAG_VXLAN_UPDATE')
                        self.EncodePbVxlanParams(msgHdr, params)
                    self.__sendmsg__(msgHdr.SerializeToString())
                    try :
                        (op, params) = self.mrouteDataSendQ.get(True, 0.001)
                    except queue.Empty:
                        queueEmpty = True
        self.log_debug("Finished execution of the thread MrouteDataSend")

    def ClagZebraListen(self):
        self.log_verbose("Executing the ClagZebra...............")
        try:
            self.ClagZebraSrvSock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.ClagZebraSrvSock.setblocking(0)
        except socket.error as msg:
            self.ClagZebraSrvSock = None
        else:
            try:
                os.remove(mroutesync._ClagZebraSocketFile)
            except OSError:
                pass
            try:
                self.ClagZebraSrvSock.bind(mroutesync._ClagZebraSocketFile)
                os.chmod(mroutesync._ClagZebraSocketFile, 0o777)
                self.ClagZebraSrvSock.listen(5)
                self.log_debug("Server Socket %r Created" % self.ClagZebraSrvSock)
            except socket.error as msg:
                self.ClagZebraSrvSock.close()
                try:
                    os.remove(mroutesync._ClagZebraSocketFile)
                except:
                    pass
                self.ClagZebraSrvSock = None
                self.log_debug("Unable to bind")
                raise

        if self.ClagZebraSrvSock:
            inputSocks = [ self.ClagZebraSrvSock ]
        self.log_debug("ClagZebraSrvSock is not None ")
        while inputSocks and self.keep_going and self.ClagZebraSrvSock:
            outputSocks = []
            for sock, sockInfo in list(self.clients.copy().items()):
                if sockInfo and sockInfo.writeBuf:
                    outputSocks.append(sock)
            try:
                (readable, writable, exceptional) = select.select(inputSocks, outputSocks,\
                                 inputSocks + outputSocks, 1)
            except Exception as e:
                self.log_debug("select failed: %r" % e)

            if self.keep_going:
                for readSock in readable:
                    if readSock is self.ClagZebraSrvSock:
                        self.zebra_stale_client_sock_shut()
                        #Accept the Connection
                        self.log_debug("Accepting the connection to zebra")
                        (newSock, addr) = self.ClagZebraSrvSock.accept()
                        self.log_debug("zebra sock %r %s created" % (newSock, addr))
                        newSock.setblocking(0)
                        #Before accepting new Connection from Zebra delete any stale zebra connections
                        sockInfo  = stream_lib.StreamInfo(newSock, addr)
                        self.clients[newSock] = sockInfo
                        self.log_debug("ZebraStatus - Connected")
                        self.SetClagZebraConn(newSock, mroutesync._CLAG_ZEBRA_CONNECTED)
                        inputSocks.append(newSock)
                    else:
                        self.log_debug("Reading data from the socket")
                        sockInfo = self.clients.get(readSock, None)
                        if sockInfo:
                            # get thenumber of bytes we want to read from this socket
                            readlen = sockInfo.get_read_data_len()
                            try:
                                myMrouteData = readSock.recv(readlen)
                            except:
                                myMrouteData = 0
                            if myMrouteData:
                                myMrouteMsg = sockInfo.proc_read_data(myMrouteData)
                                if myMrouteMsg:
                                    self.log_debug("Length of myMrouteMsg after parsing the payload is %d" % \
                                                (len(myMrouteMsg)))
                                    self.mrouteDataRecvQ.put(myMrouteMsg)
                            else:
                                if readSock in self.clagZebraConnDB:
                                    self.log_debug("ZebraStatus - Disconnected")
                                    self.SetClagZebraConn(readSock, mroutesync._CLAG_ZEBRA_DISCONNECTED)
                                if readSock in inputSocks:
                                    self.log_debug("Disconnecting the socket")
                                    inputSocks.remove(readSock)
                                self.sock_shut(readSock)

                for writeSock in writable:
                    self.log_debug("writeSock is for loop")
                    sockInfo = self.clients.get(writeSock, None)
                    if sockInfo and sockInfo.writeBuf:
                        try:
                            numSent = writeSock.send(sockInfo.writeBuf)
                            sockInfo.remove_written_data(numSent)
                        except:
                            self.log_debug("ZebraStatus - Disconnected")
                            self.SetClagZebraConn(readSock, mroutesync._CLAG_ZEBRA_DISCONNECTED)
                            self.sock_shut(writeSock)
                for errSock in exceptional:
                    if errSock in inputSocks:
                        inputSocks.remove(errSock)
                    if errSock in self.clagZebraConnDB:
                        self.log_debug("ZebraStatus - Disconnected")
                        self.SetClagZebraConn(errSock, mroutesync._CLAG_ZEBRA_DISCONNECTED)
                    self.sock_shut(errSock)

        self.log_debug("Closing the zebra Listen Thread")
        self.sock_shut(self.ClagZebraSrvSock)
