#! /usr/bin/python3
#-------------------------------------------------------------------------------
#
# Copyright 2020 Cumulus Networks, inc  all rights reserved
# Source: https://gist.github.com/niccokunzmann/6038331
#
#-------------------------------------------------------------------------------
#

try:
    import linecache
    import sys
    import threading
    import time
except ImportError as e:
    raise ImportError (str(e) + "- required module not found")

#-------------------------------------------------------------------------------
# Thread debugging tool


class StoppableThread(threading.Thread):
    """Thread class with a stop() method.

    The thread itself has to check regularly for the is_stopped()
    condition.
    """
    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stopped = False

    def stop(self):
        self._stopped = True

    def is_stopped(self):
        return self._stopped


class ThreadMonitor():

    SECONDS_FROZEN = 10  # seconds
    TEST_INTERVAL = 100  # milliseconds

    def __init__(self, Log):
        self.Log = Log

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_THREAD_MONITOR, logStr)

    def start_monitoring(self, seconds_frozen=SECONDS_FROZEN,
                     test_interval=TEST_INTERVAL):
        """Start monitoring for hanging threads.
        seconds_frozen - How much time should thread hang to activate
        printing stack trace - default(10)
        tests_interval - Sleep time of monitoring thread (in milliseconds)
        - default(100)
        """

        thread = StoppableThread(target=self.monitor, args=(seconds_frozen,
                                                   test_interval))
        thread.daemon = True
        thread.start()
        return thread

    def monitor(self, seconds_frozen, test_interval):
        """Monitoring thread function.
        Checks if thread is hanging for time defined by
        ``seconds_frozen`` parameter every ``test_interval`` milliseconds.
        """
        current_thread = threading.current_thread()
        hanging_threads = set()
        old_threads = {}  # Threads found on previous iteration.

        while not current_thread.is_stopped():
            new_threads = self.get_current_frames()

            # Report died threads.
            for thread_id, thread_data in list(old_threads.items()):
                if thread_id not in new_threads and thread_id in hanging_threads:
                    self.log_died_thread(thread_data)

            # Process live threads.
            time.sleep(test_interval/1000.)
            now = time.time()
            then = now - seconds_frozen
            for thread_id, thread_data in list(new_threads.items()):
                # Don't report the monitor thread.
                if thread_id == current_thread.ident:
                    continue
                frame = thread_data['frame']
                # If thread is new or it's stack is changed then update time.
                if (thread_id not in old_threads or
                        frame != old_threads[thread_id]['frame']):
                    thread_data['time'] = now
                    # If the thread was hanging then report awaked thread.
                    if thread_id in hanging_threads:
                        hanging_threads.remove(thread_id)
                        self.log_awaked_thread(thread_data)
                else:
                    # If stack is not changed then keep old time.
                    last_change_time = old_threads[thread_id]['time']
                    thread_data['time'] = last_change_time
                    # Check if this is a new hanging thread.
                    if (thread_id not in hanging_threads and
                            last_change_time < then):
                        hanging_threads.add(thread_id)
                        # Report the hanged thread.
                        self.log_hanged_thread(thread_data, frame)
            old_threads = new_threads

    def get_current_frames(self):
        """Return current threads prepared for
        further processing.
        """
        threads = {thread.ident: thread for thread in threading.enumerate()}
        return dict(
            (thread_id, {
                    'frame': self.thread2list(frame),
                    'time': None,
                    'id': thread_id,
                    'name': threads[thread_id].name,
                    'object': threads[thread_id]
                })
            for thread_id, frame in list(sys._current_frames().items())
        )

    def frame2string(self, frame):
        """Return info about frame.
        Keyword arg:
            frame
        Return string in format:
        File {file name}, line {line number}, in
        {name of parent of code object} {newline}
        Line from file at line number
        """
        lineno = frame.f_lineno  # or f_lasti
        co = frame.f_code
        filename = co.co_filename
        name = co.co_name
        s = '\tFile "{0}", line {1}, in {2}'.format(filename, lineno, name)
        line = linecache.getline(filename, lineno, frame.f_globals).lstrip()
        return s + '\n\t\t' + line

    def thread2list(self, frame):
        """Return list with string frame representation of each frame of
        thread.
        """
        l = []
        while frame:
            l.insert(0, self.frame2string(frame))
            frame = frame.f_back
        return l

    def threadcaption(self, thread_data):
        return 'Thread {id} "{name}"'.format(**thread_data)

    def log_hanged_thread(self, thread_data, frame):
        """Print the stack trace of the deadlock after hanging
        `seconds_frozen`.
        """
        self.write_log('{0} hangs '.format(self.threadcaption(thread_data)), ''.join(frame))

    def log_awaked_thread(self, thread_data):
        """Print message about awaked thread that was considered as
        hanging.
        """
        self.write_log('{0} awaked'.format(self.threadcaption(thread_data)))

    def log_died_thread(self, thread_data):
        """Print message about died thread that was considered as
        hanging.
        """
        self.write_log('{0} died  '.format(self.threadcaption(thread_data)))

    def write_log(self, title, message=''):
        """Write formatted log message to stderr."""
        self.log_debug(''.join([
            title.center(40).center(60, '-'), '\n', message
        ]))

