#! /usr/bin/python3
#-------------------------------------------------------------------------------
#
# Copyright 2014, 2020 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.datasync
    import clag.util as util
    import copy
    import threading
    import json
    import xml.etree.ElementTree as ET
    import clag.clagdPeerXchange_pb2 as peerXchange
except ImportError as e:
    raise ImportError (str(e) + "- required module not found")

class vlansync(clag.datasync.datasync):
    """
    This class handles everything required to send and receive the VLAN information
    between two CLAG peers.
    """

    # Max entries synced to peer in a single protobuf message
    _MAX_SYNC = 25

    def __init__(self, Intf, Parser, Log, peerSendQueue):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(vlansync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Intf   = Intf
        self.Parser = Parser
        self.Log    = Log
        self.peerSendQueue = peerSendQueue

        #
        #   VLAN Database - There are two databases, one for our switch and one
        #   for the peer switch. The structure of each data base is the same: 
        #   A dictionary which contains the dual
        #   connected interface names as the keys, and a list of 128 32-bit 
        #   numbers. The list contains 4096 bits, one for each possible VLAN ID. 
        #   The first element in the list represents VLAN IDs 31-0 with MSB
        #   VLAN ID 31 and the LSB VLAN ID 0. 
        #   {
        #       "bond3" : [ 0x00000001, 0x00010013, ... ]
        #       "bond2" : [ 0x00010F00, 0x00A00700, ... ]
        #   }
        #
        self.ourVlanDBLock  = threading.Lock()
        self.ourVlanDB      = {}
        self.peerVlanDBLock = threading.Lock()
        self.peerVlanDB     = {}
        self.peerReplyEvent = clag.clagthread.Event()

    def __del__(self):
        """
        Destructor:
        """
        pass

    def SignalThreads(self):
        self.peerReplyEvent.set()

    def JoinThreads(self):
        pass

    def GetObjectSize(self):
        vlanObj = {}
        self.ourVlanDBLock.acquire()
        vlanObj['ourVlanDB'] = util.getSize(self.ourVlanDB)
        self.ourVlanDBLock.release()
        self.peerVlanDBLock.acquire()
        vlanObj['peerVlanDB'] = util.getSize(self.peerVlanDB)
        self.peerVlanDBLock.release()
        return vlanObj

    def GetVlanIdList(self, vlanMapsList):
        vlanBitConvert = {
            0x00000001 :  0, 0x00000003 :  1, 0x00000007 :  2, 0x0000000F :  3,
            0x0000001F :  4, 0x0000003F :  5, 0x0000007F :  6, 0x000000FF :  7,
            0x000001FF :  8, 0x000003FF :  9, 0x000007FF : 10, 0x00000FFF : 11,
            0x00001FFF : 12, 0x00003FFF : 13, 0x00007FFF : 14, 0x0000FFFF : 15,
            0x0001FFFF : 16, 0x0003FFFF : 17, 0x0007FFFF : 18, 0x000FFFFF : 19,
            0x001FFFFF : 20, 0x003FFFFF : 21, 0x007FFFFF : 22, 0x00FFFFFF : 23,
            0x01FFFFFF : 24, 0x03FFFFFF : 25, 0x07FFFFFF : 26, 0x0FFFFFFF : 27,
            0x1FFFFFFF : 28, 0x3FFFFFFF : 29, 0x7FFFFFFF : 30, 0xFFFFFFFF : 31
        }
        vlanIds = []
        vlanBase = 0
        for vlanMap in vlanMapsList:
            while vlanMap:
                vlanIds.append(vlanBase + vlanBitConvert[vlanMap ^ (vlanMap - 1)])
                vlanMap &= vlanMap - 1
            vlanBase += 32
        return vlanIds
        
    def __str__(self):
        """
        Return local information in human-readable form
        """
        vlanDb = self.GetOurVlanDB()
        if vlanDb is not None:
            vlanDb = self.GetFormattedVlanDB(vlanDb)
        return json.dumps(vlanDb, sort_keys=True, indent=4)

    def EncodedOurStr(self):
        return self.__str__()

    def EncodedPeerStr(self):
        """
        Return peer information in human-readable form
        """
        vlanDb = self.GetPeerVlanDB()
        vlanDb = self.GetFormattedVlanDB(vlanDb)
        return json.dumps(vlanDb, sort_keys=True, indent=4)

    def GetOurVlanDB(self, inclPeerlink=True):
        """
        Return a copy of the current VLAN data base on this switch.
        """
        self.FillOurVlanDB(inclPeerlink)
        self.ourVlanDBLock.acquire()
        vlanDB = copy.deepcopy(self.ourVlanDB)
        self.ourVlanDBLock.release()
        return vlanDB

    def GetFormattedVlanDB(self, vlanDB):
        """
        Return a copy of formatted current VLAN data base 
        """
        formatvlanDB = {}
        for iface in vlanDB:
            formatvlanDB[iface] = self.GetVlanIdList(vlanDB[iface])
        return formatvlanDB

    def GetPeerVlanDB(self):
        """
        Return a copy of the current VLAN data base on the peer switch.
        """
        vlanDbOp = self.FillPeerVlanDB if self.Intf.isPeerAlive() else self.ClearPeerInfo
        vlanDbOp()
        self.peerVlanDBLock.acquire()
        vlanDB = copy.deepcopy(self.peerVlanDB)
        self.peerVlanDBLock.release()
        return vlanDB

    def SetOurVlanDB(self, vlanDB):
        """
        Set the current value of the VLAN data base on this switch.
        """
        self.ourVlanDBLock.acquire()
        if self.ourVlanDB != vlanDB:
            self.ourVlanDB = copy.deepcopy(vlanDB)
        self.ourVlanDBLock.release()

    def UpdatePeerVlanDB(self, vlanDB):
        self.peerVlanDBLock.acquire()
        self.peerVlanDB.update(vlanDB)
        self.peerVlanDBLock.release()

    def SetPeerVlanDB(self, vlanDB):
        """
        Return a copy of the current VLAN data base on the peer switch.
        """
        self.peerVlanDBLock.acquire()
        if self.peerVlanDB != vlanDB:
            self.peerVlanDB = copy.deepcopy(vlanDB)
        self.peerVlanDBLock.release()

    def CollectLocalInfo(self, nlm):
        """
        This function is called on a periodic basis to gather up the 
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.

        Since the VLAN information is only collected on an "as needed" basis,
        this function does nothing.
        """
        pass

    def QueueMessage(self, protobuf, msgData):
        if protobuf:
            return self.QueueProtoBufMessage(msgData)
        else:
            return self.QueueXmlMessage(msgData)

    def QueueXmlMessage(self, msgData):
        xmlStr = ET.tostring(msgData)
        self.peerSendQueue.put((None, xmlStr))
        msgData = ET.Element("clag_data", {'version': self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
        return msgData

    def QueueProtoBufMessage(self, msgData):
        if not msgData.ByteSize():
            return peerXchange.Vlans()
        self.Log.log_debug(self.Log._LOG_DEBUG_INFORM_PEER, "TXing vlan data: %s" % str(msgData))
        msgHdr = peerXchange.Header()
        msgHdr.type = peerXchange.Header.MessageType.Value('CLAG_VLAN_SYNC')
        msgHdr.data = msgData.SerializeToString()
        self.peerSendQueue.put((msgHdr, None))
        msgData = peerXchange.Vlans()
        return msgData

    def FillOurVlanDB(self, inclPeerlink=False):
        """
        This function is called to fill in the local VLAN database.
        """
        # Get the current list of dual connected hosts
        if inclPeerlink:
            peerIntf = self.Intf.GetLogicalMastersOfSubIf(self.Parser.args.peerIf)
            if not peerIntf:
                peerIntf = [self.Parser.args.peerIf]
            peerIntf = list(peerIntf)
            allIntfs = list(self.Parser.GetClagIntfDB().keys()) + peerIntf
        else:
            allIntfs = self.Intf.GetDualIfs() 

        vlanMap = self.Intf.GetMasterIfVlanMap()
        vlanDB = {}
        for intf in allIntfs:
            # Depending on the bridge model, the VLAN IDs are obtained in one of 
            # two ways. First, try the new bridge model:
            vlanDB[intf] = self.Intf.GetIfVlanMap(intf)
            if len(vlanDB[intf]) == 0:
                vlanDB[intf] = [0x00000000] * 128
                for vlanId in vlanMap.get(intf, {}):
                    vlanDB[intf][vlanId // 32] |= 1 << (vlanId % 32)
        self.SetOurVlanDB(vlanDB)

    def MakeProtoBufMessage(self):
        pass

    def FillPeerVlanDB(self):
        '''
        In order to fill the peer's VLAN database, we send a request to the peer
        and wait for him to reply.
        '''
        self.peerReplyEvent.clear()
        self.SendRequestToPeer()
        timeout = not self.peerReplyEvent.wait(6)
        if timeout:
            self.ClearPeerInfo()

    def GetIntfVlanDB(self):
        """
        Returns the map for allowed vlans on the peerlink, clag bonds
        """
        vlanDB = self.GetOurVlanDB(True)
        if vlanDB is not None:
            vlanDB = self.GetFormattedVlanDB(vlanDB)
        return vlanDB

    def SendRequestToPeer(self):
        '''
        This function sends a request to the peer to return its VLAN information.
        '''
        protobuf = self.Parser.peerProtoBuf
        if protobuf:
            msgData = peerXchange.Vlans()
            msgData.vlanRequest = True
            self.ClearPeerInfo()
        else:
            msgData = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
            ET.SubElement(msgData, "vlan-request")
        self.QueueMessage(protobuf, msgData)

    def SendInfoToPeer(self):
        """
        Writes the local information as a protocol buffer message.
        """
        vlanEntries = 0
        vlanDB = self.GetOurVlanDB()
        msgData = peerXchange.Vlans()
        for bond in vlanDB:
            msgDataVlan = msgData.vlan.add()
            vlanMapStr = ""
            for vlanMap in vlanDB[bond]:
                vlanMapStr += "0x%08x " % (vlanMap,)
            msgDataVlan.dualIf = bond
            msgDataVlan.vlanMap = vlanMapStr
            vlanEntries += 1
            if vlanEntries >= vlansync._MAX_SYNC:
                msgData = self.QueueProtoBufMessage(msgData)
                vlanEntries = 0
        msgData.vlanReplyDone = True
        self.QueueProtoBufMessage(msgData)

    def SendXmlInfoToPeer(self):
        """
        Encodes the local information as XML.

        parent - An xml.etree.ElementTree.Element instance which is the parent
            element under which all of the information should be encoded.

        <vlans>
            <vlanmap dualIf="bond4">
                0x00000100 0x02D04002 0xD3000145 ...
            </vlanmap>
            <vlanmap dualIf="bond5">
                0x00000100 0x02D04002 0xD3000145 ...
            </vlanmap>
        </vlans>
        """
        vlanDB = self.GetOurVlanDB()
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion, 'packageVersion' : self.Parser.packageVersion})
        vlans = ET.SubElement(root, "vlans")
        for bond in vlanDB:
            vlanMapStr = ""
            for vlanMap in vlanDB[bond]:
                vlanMapStr += "0x%08x " % (vlanMap,)
            vlanAttrib = { "dualIf" : bond }
            ET.SubElement(vlans, "vlanmap", vlanAttrib).text = vlanMapStr
        self.QueueXmlMessage(root)

    def ParseProtoBufMessage(self, msgHdr):
        """
        Reads the protobuf information sent from the peer.
        """
        msgData = peerXchange.Vlans()
        msgData.ParseFromString(msgHdr.data)

        if msgData.vlanRequest:
            self.SendInfoToPeer()
            return msgData

        vlanDB = {}
        for vlan in msgData.vlan:
            dualIf = vlan.dualIf
            vlanMapStr = str(vlan.vlanMap)
            vlanDB[dualIf] = [ int(i,0) for i in vlanMapStr.split() ]
        self.UpdatePeerVlanDB(vlanDB)
        if msgData.vlanReplyDone:
            self.peerReplyEvent.set()
        return msgData

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.

        root - The root element of the XML tree
        """
        vlans = root.find('vlans')
        if vlans is not None:
            vlanDB = {}
            for vlanmap in vlans.findall('vlanmap'):
                dualIf = vlanmap.get('dualIf')
                vlanDB[dualIf] = [ int(i,0) for i in vlanmap.text.split() ]
            self.SetPeerVlanDB(vlanDB)
            self.peerReplyEvent.set()
            if not self.Parser.isPeerVlanSyncOptimized():
                self.SendXmlInfoToPeer()
        vlanrequest = root.find('vlan-request')
        if vlanrequest is not None:
            self.SendXmlInfoToPeer()


    def DualIfChange(self, dualIf, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        pass

    def MatchIdIfChange(self, dualIf, enable):
        """
        Called when an interface gets or loses a matching clag-id on the peer
        """
        pass

    def ClearPeerInfo(self):
        """
        Removes information which has been obtained from the peer.
        """
        self.SetPeerVlanDB({})

    def ClearLocalInfo(self):
        """
        Removes information which has been obtained locally.
        """
        self.SetOurVlanDB({})
